import json
from typing import Dict, Optional, Union
import httpx

from unbound.api_resources.types.chat_complete_type import (
    ChatCompletionMessage,
    Choice,
    StreamChoice,
    Usage,
)
from unbound.api_resources.types.complete_type import Logprobs, TextChoice

from .utils import parse_headers
from typing import List, Any
from pydantic import BaseModel, PrivateAttr


class PromptCompletion(BaseModel):
    id: Optional[str] = None
    choices: Optional[List[Choice]] = None
    created: Optional[int] = None
    model: Optional[str] = None
    object: Optional[str] = None
    system_fingerprint: Optional[str] = None
    usage: Optional[Usage] = None
    index: Optional[int] = None
    text: Optional[str] = None
    logprobs: Optional[Logprobs] = None
    finish_reason: Optional[str] = None
    _headers: Optional[httpx.Headers] = PrivateAttr()

    def __str__(self):
        return json.dumps(self.dict(), indent=4)

    def __getitem__(self, key):
        return getattr(self, key, None)

    def get(self, key: str, default: Optional[Any] = None):
        return getattr(self, key, None) or default

    def get_headers(self) -> Optional[Dict[str, str]]:
        return parse_headers(self._headers)


class PromptCompletionChunk(BaseModel):
    id: Optional[str] = None
    object: Optional[str] = None
    created: Optional[int] = None
    model: Optional[str] = None
    provider: Optional[str] = None
    choices: Optional[Union[List[TextChoice], List[StreamChoice]]] = None

    def __str__(self):
        return json.dumps(self.dict(), indent=4)

    def __getitem__(self, key):
        return getattr(self, key, None)

    def get(self, key: str, default: Optional[Any] = None):
        return getattr(self, key, None) or default


FunctionParameters = Dict[str, object]


class Function(BaseModel):
    name: Optional[str]
    description: Optional[str] = None
    parameters: Optional[FunctionParameters] = None


class Tool(BaseModel):
    function: Function
    type: Optional[str]


class PromptRenderData(BaseModel):
    messages: Optional[List[ChatCompletionMessage]] = None
    prompt: Optional[str] = None
    model: Optional[str] = None
    suffix: Optional[str] = None
    max_tokens: Optional[int] = None
    temperature: Optional[float] = None
    top_k: Optional[int] = None
    top_p: Optional[float] = None
    n: Optional[int] = None
    stop_sequences: Optional[List[str]] = None
    timeout: Union[float, None] = None
    functions: Optional[List[Function]] = None
    function_call: Optional[Union[None, str, Function]] = None
    logprobs: Optional[bool] = None
    top_logprobs: Optional[int] = None
    echo: Optional[bool] = None
    stop: Optional[Union[str, List[str]]] = None
    presence_penalty: Optional[int] = None
    frequency_penalty: Optional[int] = None
    best_of: Optional[int] = None
    logit_bias: Optional[Dict[str, int]] = None
    user: Optional[str] = None
    organization: Optional[str] = None
    tool_choice: Optional[Union[None, str]] = None
    tools: Optional[List[Tool]] = None


class PromptRender(BaseModel):
    success: Optional[bool] = True
    data: PromptRenderData

    def __str__(self):
        return json.dumps(self.dict(), indent=4)

    def __getitem__(self, key):
        return getattr(self, key, None)

    def get(self, key: str, default: Optional[Any] = None):
        return getattr(self, key, None) or default



from typing import Optional

import httpx

from unbound.api_resources.global_constants import UNBOUND_HEADER_PREFIX


def parse_headers(headers: Optional[httpx.Headers]) -> dict:
    if headers is None:
        return {}

    _headers = {}
    for k, v in headers.items():
        if k.startswith(UNBOUND_HEADER_PREFIX):
            k = k.replace(UNBOUND_HEADER_PREFIX, "")
            _headers[k] = v

    return _headers



from __future__ import annotations
import asyncio

import json
from types import TracebackType
from typing import (
    Dict,
    Any,
    List,
    Union,
    Mapping,
    cast,
    Optional,
    Type,
    overload,
    Literal,
    get_args,
)
import httpx
import platform

from unbound.api_resources.apis.create_headers import createHeaders
from .global_constants import (
    UNBOUND_HEADER_PREFIX,
)
from .utils import remove_empty_values, Options, set_base_url
from .exceptions import (
    APIStatusError,
    APITimeoutError,
    APIConnectionError,
)
from unbound.version import VERSION
from .utils import ResponseT, make_status_error, default_api_key
from .common_types import StreamT, AsyncStreamT
from .streaming import Stream, AsyncStream


class MissingStreamClassError(TypeError):
    def __init__(self) -> None:
        super().__init__(
            "The `stream` argument was set to `True` but the `stream_cls` argument was\
            not given",
        )


class APIClient:
    _client: httpx.Client
    _default_stream_cls: Union[type[Stream[Any]], None] = None

    def __init__(
        self,
        *,
        base_url: Optional[str] = None,
        api_key: Optional[str] = None,
        virtual_key: Optional[str] = None,
        config: Optional[Union[Mapping, str]] = None,
        provider: Optional[str] = None,
        trace_id: Optional[str] = None,
        metadata: Union[Optional[dict[str, str]], str] = None,
        cache_namespace: Optional[str] = None,
        debug: Optional[bool] = None,
        cache_force_refresh: Optional[bool] = None,
        custom_host: Optional[str] = None,
        forward_headers: Optional[List[str]] = None,
        openai_project: Optional[str] = None,
        openai_organization: Optional[str] = None,
        aws_secret_access_key: Optional[str] = None,
        aws_access_key_id: Optional[str] = None,
        aws_session_token: Optional[str] = None,
        aws_region: Optional[str] = None,
        vertex_project_id: Optional[str] = None,
        vertex_region: Optional[str] = None,
        workers_ai_account_id: Optional[str] = None,
        azure_resource_name: Optional[str] = None,
        azure_deployment_id: Optional[str] = None,
        azure_api_version: Optional[str] = None,
        azure_endpoint_name: Optional[str] = None,
        huggingface_base_url: Optional[str] = None,
        http_client: Optional[httpx.Client] = None,
        request_timeout: Optional[int] = None,
        strict_open_ai_compliance: Optional[bool] = None,
        anthropic_beta: Optional[str] = None,
        anthropic_version: Optional[str] = None,
        mistral_fim_completion: Optional[str] = None,
        **kwargs,
    ) -> None:
        self.base_url = set_base_url(base_url, api_key)
        self.api_key = default_api_key(self.base_url, api_key)
        self.virtual_key = virtual_key
        self.config = config
        self.provider = provider
        self.trace_id = trace_id
        self.metadata = metadata
        self.debug = debug
        self.cache_force_refresh = cache_force_refresh
        self.custom_host = custom_host
        self.forward_headers = forward_headers
        self.openai_project = openai_project
        self.openai_organization = openai_organization
        self.aws_secret_access_key = aws_secret_access_key
        self.aws_access_key_id = aws_access_key_id
        self.aws_session_token = aws_session_token
        self.aws_region = aws_region
        self.vertex_project_id = vertex_project_id
        self.vertex_region = vertex_region
        self.workers_ai_account_id = workers_ai_account_id
        self.azure_resource_name = azure_resource_name
        self.azure_deployment_id = azure_deployment_id
        self.azure_api_version = azure_api_version
        self.azure_endpoint_name = azure_endpoint_name
        self.huggingface_base_url = huggingface_base_url
        self.cache_namespace = cache_namespace
        self.request_timeout = request_timeout
        self.strict_open_ai_compliance = strict_open_ai_compliance
        self.anthropic_beta = anthropic_beta
        self.anthropic_version = anthropic_version
        self.mistral_fim_completion = mistral_fim_completion
        self.kwargs = kwargs

        self.custom_headers = createHeaders(
            api_key=api_key,
            virtual_key=virtual_key,
            config=config,
            provider=provider,
            trace_id=trace_id,
            metadata=metadata,
            debug=debug,
            cache_force_refresh=cache_force_refresh,
            custom_host=custom_host,
            forward_headers=forward_headers,
            openai_project=openai_project,
            openai_organization=openai_organization,
            aws_secret_access_key=aws_secret_access_key,
            aws_access_key_id=aws_access_key_id,
            aws_session_token=aws_session_token,
            aws_region=aws_region,
            vertex_project_id=vertex_project_id,
            vertex_region=vertex_region,
            workers_ai_account_id=workers_ai_account_id,
            azure_resource_name=azure_resource_name,
            azure_deployment_id=azure_deployment_id,
            azure_api_version=azure_api_version,
            azure_endpoint_name=azure_endpoint_name,
            huggingface_base_url=huggingface_base_url,
            cache_namespace=cache_namespace,
            request_timeout=request_timeout,
            strict_open_ai_compliance=strict_open_ai_compliance,
            anthropic_beta=anthropic_beta,
            anthropic_version=anthropic_version,
            mistral_fim_completion=mistral_fim_completion,
            **kwargs,
        )

        self.allHeaders = self._build_headers(Options.construct())
        self._client = http_client or httpx.Client(
            base_url=self.base_url,
            headers={
                "Accept": "application/json",
            },
        )

        self.response_headers: httpx.Headers | None = None

    def _serialize_header_values(
        self, headers: Optional[Mapping[str, Any]]
    ) -> Dict[str, str]:
        if headers is None:
            return {}
        return {
            f"{UNBOUND_HEADER_PREFIX}{k}": json.dumps(v)
            if isinstance(v, (dict, list))
            else str(v)
            for k, v in headers.items()
        }

    @property
    def custom_auth(self) -> Optional[httpx.Auth]:
        return None

    @overload
    def _post(
        self,
        path: str,
        *,
        body: Mapping[str, Any] = {},
        files: Any = None,
        cast_to: Type[ResponseT],
        stream: Literal[True],
        stream_cls: type[StreamT],
        params: Mapping[str, str],
        headers: Mapping[str, str],
    ) -> StreamT:
        ...

    @overload
    def _post(
        self,
        path: str,
        *,
        body: Mapping[str, Any] = {},
        files: Any = None,
        cast_to: Type[ResponseT],
        stream: Literal[False],
        stream_cls: type[StreamT],
        params: Mapping[str, str],
        headers: Mapping[str, str],
    ) -> ResponseT:
        ...

    @overload
    def _post(
        self,
        path: str,
        *,
        body: Mapping[str, Any] = {},
        files: Any = None,
        cast_to: Type[ResponseT],
        stream: bool,
        stream_cls: type[StreamT],
        params: Mapping[str, str],
        headers: Mapping[str, str],
    ) -> Union[ResponseT, StreamT]:
        ...

    def _post(
        self,
        path: str,
        *,
        body: Mapping[str, Any] = {},
        files: Any = None,
        cast_to: Type[ResponseT],
        stream: bool,
        stream_cls: type[StreamT],
        params: Mapping[str, str],
        headers: Mapping[str, str],
    ) -> Union[ResponseT, StreamT]:
        if path.endswith("/generate"):
            opts = self._construct_generate_options(
                method="post",
                url=path,
                body=body,
                files=files,
                stream=stream,
                params=params,
                headers=headers,
            )
        else:
            opts = self._construct(
                method="post",
                url=path,
                body=body,
                files=files,
                stream=stream,
                params=params,
                headers=headers,
            )

        res = self._request(
            options=opts,
            stream=stream,
            cast_to=cast_to,
            stream_cls=stream_cls,
        )
        return res

    @overload
    def _put(
        self,
        path: str,
        *,
        body: Mapping[str, Any],
        cast_to: Type[ResponseT],
        stream: Literal[True],
        stream_cls: type[StreamT],
        params: Mapping[str, str],
        headers: Mapping[str, str],
    ) -> StreamT:
        ...

    @overload
    def _put(
        self,
        path: str,
        *,
        body: Mapping[str, Any],
        cast_to: Type[ResponseT],
        stream: Literal[False],
        stream_cls: type[StreamT],
        params: Mapping[str, str],
        headers: Mapping[str, str],
    ) -> ResponseT:
        ...

    @overload
    def _put(
        self,
        path: str,
        *,
        body: Mapping[str, Any],
        cast_to: Type[ResponseT],
        stream: bool,
        stream_cls: type[StreamT],
        params: Mapping[str, str],
        headers: Mapping[str, str],
    ) -> Union[ResponseT, StreamT]:
        ...

    def _put(
        self,
        path: str,
        *,
        body: Mapping[str, Any],
        cast_to: Type[ResponseT],
        stream: bool,
        stream_cls: type[StreamT],
        params: Mapping[str, str],
        headers: Mapping[str, str],
    ) -> Union[ResponseT, StreamT]:
        opts = self._construct(
            method="put",
            url=path,
            body=body,
            stream=stream,
            params=params,
            headers=headers,
        )
        res = self._request(
            options=opts,
            stream=stream,
            cast_to=cast_to,
            stream_cls=stream_cls,
        )
        return res

    @overload
    def _get(
        self,
        path: str,
        *,
        body: Mapping[str, Any],
        params: Mapping[str, str],
        headers: Mapping[str, str],
        cast_to: Type[ResponseT],
        stream: Literal[True],
        stream_cls: type[StreamT],
    ) -> StreamT:
        ...

    @overload
    def _get(
        self,
        path: str,
        *,
        body: Mapping[str, Any],
        params: Mapping[str, str],
        headers: Mapping[str, str],
        cast_to: Type[ResponseT],
        stream: Literal[False],
        stream_cls: type[StreamT],
    ) -> ResponseT:
        ...

    @overload
    def _get(
        self,
        path: str,
        *,
        body: Mapping[str, Any],
        params: Mapping[str, str],
        headers: Mapping[str, str],
        cast_to: Type[ResponseT],
        stream: bool,
        stream_cls: type[StreamT],
    ) -> Union[ResponseT, StreamT]:
        ...

    def _get(
        self,
        path: str,
        *,
        body: Mapping[str, Any],
        params: Mapping[str, str],
        headers: Mapping[str, str],
        cast_to: Type[ResponseT],
        stream: bool,
        stream_cls: type[StreamT],
    ) -> Union[ResponseT, StreamT]:
        opts = self._construct(
            method="get",
            url=path,
            body=body,
            stream=stream,
            params=params,
            headers=headers,
        )
        res = self._request(
            options=opts,
            stream=stream,
            cast_to=cast_to,
            stream_cls=stream_cls,
        )
        return res

    @overload
    def _delete(
        self,
        path: str,
        *,
        body: Mapping[str, Any],
        params: Mapping[str, str],
        headers: Mapping[str, str],
        cast_to: Type[ResponseT],
        stream: Literal[True],
        stream_cls: type[StreamT],
    ) -> StreamT:
        ...

    @overload
    def _delete(
        self,
        path: str,
        *,
        body: Mapping[str, Any],
        params: Mapping[str, str],
        headers: Mapping[str, str],
        cast_to: Type[ResponseT],
        stream: Literal[False],
        stream_cls: type[StreamT],
    ) -> ResponseT:
        ...

    @overload
    def _delete(
        self,
        path: str,
        *,
        body: Mapping[str, Any],
        params: Mapping[str, str],
        headers: Mapping[str, str],
        cast_to: Type[ResponseT],
        stream: bool,
        stream_cls: type[StreamT],
    ) -> Union[ResponseT, StreamT]:
        ...

    def _delete(
        self,
        path: str,
        *,
        body: Mapping[str, Any],
        params: Mapping[str, str],
        headers: Mapping[str, str],
        cast_to: Type[ResponseT],
        stream: bool,
        stream_cls: type[StreamT],
    ) -> Union[ResponseT, StreamT]:
        opts = self._construct(
            method="delete",
            url=path,
            body=body,
            stream=stream,
            params=params,
            headers=headers,
        )
        res = self._request(
            options=opts,
            stream=stream,
            cast_to=cast_to,
            stream_cls=stream_cls,
        )
        return res

    def _construct_generate_options(
        self,
        *,
        method: str,
        url: str,
        body: Any,
        files: Any,
        stream: bool,
        params: Mapping[str, str],
        headers: Mapping[str, str],
    ) -> Options:
        opts = Options.construct()
        opts.method = method
        opts.url = url
        json_body = body
        opts.files = files
        opts.json_body = remove_empty_values(json_body)
        opts.headers = remove_empty_values(headers)
        return opts

    def _construct(
        self,
        *,
        method: str,
        url: str,
        body: Mapping[str, Any],
        files: Any = None,
        stream: bool,
        params: Mapping[str, str],
        headers: Mapping[str, str],
    ) -> Options:
        opts = Options.construct()
        opts.method = method
        opts.url = url
        opts.files = files
        if method != "get":
            opts.json_body = remove_empty_values(body)
        opts.headers = remove_empty_values(headers)
        return opts

    @property
    def _default_headers(self) -> Mapping[str, str]:
        return {
            f"{UNBOUND_HEADER_PREFIX}api-key": self.api_key,
            f"{UNBOUND_HEADER_PREFIX}package-version": f"unbound-{VERSION}",
            f"{UNBOUND_HEADER_PREFIX}runtime": platform.python_implementation(),
            f"{UNBOUND_HEADER_PREFIX}runtime-version": platform.python_version(),
        }

    def _build_headers(self, options: Options) -> Dict[str, Any]:
        option_headers = options.headers or {}
        headers_dict = self._merge_mappings(
            self._default_headers, option_headers, self.custom_headers
        )
        return headers_dict

    def _merge_mappings(
        self,
        *args,
    ) -> Dict[str, Any]:
        """Merge two mappings of the given type
        In cases with duplicate keys the second mapping takes precedence.
        """
        mapped_headers = {}
        for i in args:
            mapped_headers.update(i)
        return mapped_headers

    def is_closed(self) -> bool:
        return self._client.is_closed

    def close(self) -> None:
        """Close the underlying HTTPX client.

        The client will *not* be usable after this.
        """
        if hasattr(self, "_client"):
            self._client.close()

    def __enter__(self: Any) -> Any:
        return self

    def __exit__(
        self,
        exc_type: Optional[BaseException],
        exc: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        self.close()

    def _build_request(self, options: Options) -> httpx.Request:
        headers = self._build_headers(options)
        params = options.params
        json_body = options.json_body
        request = self._client.build_request(
            method=options.method,
            url=options.url,
            headers=headers,
            params=params,
            json=json_body,
            files=options.files,
            timeout=options.timeout,
        )
        return request

    @overload
    def _request(
        self,
        *,
        options: Options,
        stream: Literal[False],
        cast_to: Type[ResponseT],
        stream_cls: Type[StreamT],
    ) -> ResponseT:
        ...

    @overload
    def _request(
        self,
        *,
        options: Options,
        stream: Literal[True],
        cast_to: Type[ResponseT],
        stream_cls: Type[StreamT],
    ) -> StreamT:
        ...

    @overload
    def _request(
        self,
        *,
        options: Options,
        stream: bool,
        cast_to: Type[ResponseT],
        stream_cls: Type[StreamT],
    ) -> Union[ResponseT, StreamT]:
        ...

    def _request(
        self,
        *,
        options: Options,
        stream: bool,
        cast_to: Type[ResponseT],
        stream_cls: Type[StreamT],
    ) -> Union[ResponseT, StreamT]:
        request = self._build_request(options)
        try:
            res = self._client.send(request, auth=self.custom_auth, stream=stream)
            res.raise_for_status()
        except httpx.HTTPStatusError as err:  # 4xx and 5xx errors
            # If the response is streamed then we need to explicitly read the response
            # to completion before attempting to access the response text.
            err.response.read()
            raise self._make_status_error_from_response(request, err.response) from None
        except httpx.TimeoutException as err:
            raise APITimeoutError(request=request) from err
        except Exception as err:
            raise APIConnectionError(request=request) from err

        self.response_headers = res.headers
        if stream or res.headers["content-type"] == "text/event-stream":
            if stream_cls is None:
                raise MissingStreamClassError()
            stream_response = stream_cls(
                response=res, cast_to=self._extract_stream_chunk_type(stream_cls)
            )
            return stream_response

        response = (
            cast(
                ResponseT,
                cast_to(**res.json()),
            )
            if not isinstance(cast_to, httpx.Response)
            else cast(ResponseT, res)
        )
        response._headers = res.headers  # type: ignore
        return response

    def _extract_stream_chunk_type(self, stream_cls: Type) -> type:
        args = get_args(stream_cls)
        if not args:
            raise TypeError(
                f"Expected stream_cls to have been given a generic type argument, e.g. \
                    Stream[Foo] but received {stream_cls}",
            )
        return cast(type, args[0])

    def _make_status_error_from_response(
        self,
        request: httpx.Request,
        response: httpx.Response,
    ) -> APIStatusError:
        err_text = response.text.strip()
        body = err_text

        try:
            body = json.loads(err_text)["error"]["message"]
            err_msg = f"Error code: {response.status_code} - {body}"
        except Exception:
            err_msg = err_text or f"Error code: {response.status_code}"

        return make_status_error(err_msg, body=body, request=request, response=response)


class AsyncHttpxClientWrapper(httpx.AsyncClient):
    def __del__(self) -> None:
        try:
            asyncio.get_running_loop().create_task(self.aclose())
        except Exception:
            pass


class AsyncAPIClient:
    _client: httpx.AsyncClient
    _default_stream_cls: Union[type[AsyncStream[Any]], None] = None

    def __init__(
        self,
        *,
        base_url: Optional[str] = None,
        api_key: Optional[str] = None,
        virtual_key: Optional[str] = None,
        config: Optional[Union[Mapping, str]] = None,
        provider: Optional[str] = None,
        trace_id: Optional[str] = None,
        metadata: Union[Optional[dict[str, str]], str] = None,
        cache_namespace: Optional[str] = None,
        debug: Optional[bool] = None,
        cache_force_refresh: Optional[bool] = None,
        custom_host: Optional[str] = None,
        forward_headers: Optional[List[str]] = None,
        openai_project: Optional[str] = None,
        openai_organization: Optional[str] = None,
        aws_secret_access_key: Optional[str] = None,
        aws_access_key_id: Optional[str] = None,
        aws_session_token: Optional[str] = None,
        aws_region: Optional[str] = None,
        vertex_project_id: Optional[str] = None,
        vertex_region: Optional[str] = None,
        workers_ai_account_id: Optional[str] = None,
        azure_resource_name: Optional[str] = None,
        azure_deployment_id: Optional[str] = None,
        azure_api_version: Optional[str] = None,
        azure_endpoint_name: Optional[str] = None,
        huggingface_base_url: Optional[str] = None,
        http_client: Optional[httpx.AsyncClient] = None,
        request_timeout: Optional[int] = None,
        strict_open_ai_compliance: Optional[bool] = None,
        anthropic_beta: Optional[str] = None,
        anthropic_version: Optional[str] = None,
        mistral_fim_completion: Optional[str] = None,
        **kwargs,
    ) -> None:
        self.base_url = set_base_url(base_url, api_key)
        self.api_key = default_api_key(self.base_url, api_key)
        self.virtual_key = virtual_key
        self.config = config
        self.provider = provider
        self.trace_id = trace_id
        self.metadata = metadata
        self.debug = debug
        self.cache_force_refresh = cache_force_refresh
        self.custom_host = custom_host
        self.forward_headers = forward_headers
        self.openai_project = openai_project
        self.openai_organization = openai_organization
        self.aws_secret_access_key = aws_secret_access_key
        self.aws_access_key_id = aws_access_key_id
        self.aws_session_token = aws_session_token
        self.aws_region = aws_region
        self.vertex_project_id = vertex_project_id
        self.vertex_region = vertex_region
        self.workers_ai_account_id = workers_ai_account_id
        self.azure_resource_name = azure_resource_name
        self.azure_deployment_id = azure_deployment_id
        self.azure_api_version = azure_api_version
        self.azure_endpoint_name = azure_endpoint_name
        self.huggingface_base_url = huggingface_base_url
        self.cache_namespace = cache_namespace
        self.request_timeout = request_timeout
        self.strict_open_ai_compliance = strict_open_ai_compliance
        self.anthropic_beta = anthropic_beta
        self.anthropic_version = anthropic_version
        self.mistral_fim_completion = mistral_fim_completion
        self.kwargs = kwargs

        self.custom_headers = createHeaders(
            api_key=api_key,
            virtual_key=virtual_key,
            config=config,
            provider=provider,
            trace_id=trace_id,
            metadata=metadata,
            debug=debug,
            cache_force_refresh=cache_force_refresh,
            custom_host=custom_host,
            forward_headers=forward_headers,
            openai_project=openai_project,
            openai_organization=openai_organization,
            aws_secret_access_key=aws_secret_access_key,
            aws_access_key_id=aws_access_key_id,
            aws_session_token=aws_session_token,
            aws_region=aws_region,
            vertex_project_id=vertex_project_id,
            vertex_region=vertex_region,
            workers_ai_account_id=workers_ai_account_id,
            azure_resource_name=azure_resource_name,
            azure_deployment_id=azure_deployment_id,
            azure_api_version=azure_api_version,
            azure_endpoint_name=azure_endpoint_name,
            huggingface_base_url=huggingface_base_url,
            cache_namespace=cache_namespace,
            request_timeout=request_timeout,
            strict_open_ai_compliance=strict_open_ai_compliance,
            anthropic_beta=anthropic_beta,
            anthropic_version=anthropic_version,
            mistral_fim_completion=mistral_fim_completion,
            **kwargs,
        )

        self.allHeaders = self._build_headers(Options.construct())
        self._client = http_client or AsyncHttpxClientWrapper(
            base_url=self.base_url,
            headers={
                "Accept": "application/json",
            },
        )

        self.response_headers: httpx.Headers | None = None

    def _serialize_header_values(
        self, headers: Optional[Mapping[str, Any]]
    ) -> Dict[str, str]:
        if headers is None:
            return {}
        return {
            f"{UNBOUND_HEADER_PREFIX}{k}": json.dumps(v)
            if isinstance(v, (dict, list))
            else str(v)
            for k, v in headers.items()
        }

    @property
    def custom_auth(self) -> Optional[httpx.Auth]:
        return None

    @overload
    async def _post(
        self,
        path: str,
        *,
        cast_to: Type[ResponseT],
        body: Mapping[str, Any],
        files: Any = None,
        stream: Literal[False],
        stream_cls: type[AsyncStreamT],
        params: Mapping[str, str],
        headers: Mapping[str, str],
    ) -> ResponseT:
        ...

    @overload
    async def _post(
        self,
        path: str,
        *,
        cast_to: Type[ResponseT],
        body: Mapping[str, Any],
        files: Any = None,
        stream: Literal[True],
        stream_cls: type[AsyncStreamT],
        params: Mapping[str, str],
        headers: Mapping[str, str],
    ) -> AsyncStreamT:
        ...

    @overload
    async def _post(
        self,
        path: str,
        *,
        cast_to: Type[ResponseT],
        body: Mapping[str, Any],
        files: Any = None,
        stream: bool,
        stream_cls: type[AsyncStreamT],
        params: Mapping[str, str],
        headers: Mapping[str, str],
    ) -> Union[ResponseT, AsyncStreamT]:
        ...

    async def _post(
        self,
        path: str,
        *,
        cast_to: Type[ResponseT],
        body: Mapping[str, Any],
        files: Any = None,
        stream: bool,
        stream_cls: type[AsyncStreamT],
        params: Mapping[str, str],
        headers: Mapping[str, str],
    ) -> Union[ResponseT, AsyncStreamT]:
        if path.endswith("/generate"):
            opts = await self._construct_generate_options(
                method="post",
                url=path,
                body=body,
                files=files,
                stream=stream,
                params=params,
                headers=headers,
            )
        else:
            opts = await self._construct(
                method="post",
                url=path,
                body=body,
                files=files,
                stream=stream,
                params=params,
                headers=headers,
            )

        res = await self._request(
            options=opts,
            stream=stream,
            cast_to=cast_to,
            stream_cls=stream_cls,
        )
        return res

    @overload
    async def _put(
        self,
        path: str,
        *,
        cast_to: Type[ResponseT],
        body: Mapping[str, Any],
        stream: Literal[False],
        stream_cls: type[AsyncStreamT],
        params: Mapping[str, str],
        headers: Mapping[str, str],
    ) -> ResponseT:
        ...

    @overload
    async def _put(
        self,
        path: str,
        *,
        cast_to: Type[ResponseT],
        body: Mapping[str, Any],
        stream: Literal[True],
        stream_cls: type[AsyncStreamT],
        params: Mapping[str, str],
        headers: Mapping[str, str],
    ) -> AsyncStreamT:
        ...

    @overload
    async def _put(
        self,
        path: str,
        *,
        cast_to: Type[ResponseT],
        body: Mapping[str, Any],
        stream: bool,
        stream_cls: type[AsyncStreamT],
        params: Mapping[str, str],
        headers: Mapping[str, str],
    ) -> Union[ResponseT, AsyncStreamT]:
        ...

    async def _put(
        self,
        path: str,
        *,
        cast_to: Type[ResponseT],
        body: Mapping[str, Any],
        stream: bool,
        stream_cls: type[AsyncStreamT],
        params: Mapping[str, str],
        headers: Mapping[str, str],
    ) -> Union[ResponseT, AsyncStreamT]:
        opts = await self._construct(
            method="put",
            url=path,
            body=body,
            stream=stream,
            params=params,
            headers=headers,
        )
        res = await self._request(
            options=opts,
            stream=stream,
            cast_to=cast_to,
            stream_cls=stream_cls,
        )
        return res

    @overload
    async def _get(
        self,
        path: str,
        *,
        body: Mapping[str, Any],
        params: Mapping[str, str],
        headers: Mapping[str, str],
        cast_to: Type[ResponseT],
        stream: Literal[True],
        stream_cls: type[AsyncStreamT],
    ) -> AsyncStreamT:
        ...

    @overload
    async def _get(
        self,
        path: str,
        *,
        body: Mapping[str, Any],
        params: Mapping[str, str],
        headers: Mapping[str, str],
        cast_to: Type[ResponseT],
        stream: Literal[False],
        stream_cls: type[AsyncStreamT],
    ) -> ResponseT:
        ...

    @overload
    async def _get(
        self,
        path: str,
        *,
        body: Mapping[str, Any],
        params: Mapping[str, str],
        headers: Mapping[str, str],
        cast_to: Type[ResponseT],
        stream: bool,
        stream_cls: type[AsyncStreamT],
    ) -> Union[ResponseT, AsyncStreamT]:
        ...

    async def _get(
        self,
        path: str,
        *,
        body: Mapping[str, Any],
        params: Mapping[str, str],
        headers: Mapping[str, str],
        cast_to: Type[ResponseT],
        stream: bool,
        stream_cls: type[AsyncStreamT],
    ) -> Union[ResponseT, AsyncStreamT]:
        opts = await self._construct(
            method="get",
            url=path,
            body=body,
            stream=stream,
            params=params,
            headers=headers,
        )
        res = await self._request(
            options=opts,
            stream=stream,
            cast_to=cast_to,
            stream_cls=stream_cls,
        )
        return res

    @overload
    async def _delete(
        self,
        path: str,
        *,
        body: Mapping[str, Any],
        params: Mapping[str, str],
        headers: Mapping[str, str],
        cast_to: Type[ResponseT],
        stream: Literal[True],
        stream_cls: type[AsyncStreamT],
    ) -> AsyncStreamT:
        ...

    @overload
    async def _delete(
        self,
        path: str,
        *,
        body: Mapping[str, Any],
        params: Mapping[str, str],
        headers: Mapping[str, str],
        cast_to: Type[ResponseT],
        stream: Literal[False],
        stream_cls: type[AsyncStreamT],
    ) -> ResponseT:
        ...

    @overload
    async def _delete(
        self,
        path: str,
        *,
        body: Mapping[str, Any],
        params: Mapping[str, str],
        headers: Mapping[str, str],
        cast_to: Type[ResponseT],
        stream: bool,
        stream_cls: type[AsyncStreamT],
    ) -> Union[ResponseT, AsyncStreamT]:
        ...

    async def _delete(
        self,
        path: str,
        *,
        body: Mapping[str, Any],
        params: Mapping[str, str],
        headers: Mapping[str, str],
        cast_to: Type[ResponseT],
        stream: bool,
        stream_cls: type[AsyncStreamT],
    ) -> Union[ResponseT, AsyncStreamT]:
        opts = await self._construct(
            method="delete",
            url=path,
            body=body,
            stream=stream,
            params=params,
            headers=headers,
        )
        res = await self._request(
            options=opts,
            stream=stream,
            cast_to=cast_to,
            stream_cls=stream_cls,
        )
        return res

    async def _construct_generate_options(
        self,
        *,
        method: str,
        url: str,
        body: Any,
        files: Any,
        stream: bool,
        params: Mapping[str, str],
        headers: Mapping[str, str],
    ) -> Options:
        opts = Options.construct()
        opts.method = method
        opts.url = url
        json_body = body
        opts.files = files
        opts.json_body = remove_empty_values(json_body)
        opts.headers = remove_empty_values(headers)
        return opts

    async def _construct(
        self,
        *,
        method: str,
        url: str,
        body: Mapping[str, Any],
        files: Any = None,
        stream: bool,
        params: Mapping[str, str],
        headers: Mapping[str, str],
    ) -> Options:
        opts = Options.construct()
        opts.method = method
        opts.url = url
        opts.files = files
        if method != "get" or method != "delete":
            opts.json_body = remove_empty_values(body)
        opts.headers = remove_empty_values(headers)
        return opts

    @property
    def _default_headers(self) -> Mapping[str, str]:
        return {
            f"{UNBOUND_HEADER_PREFIX}api-key": self.api_key,
            f"{UNBOUND_HEADER_PREFIX}package-version": f"unbound-{VERSION}",
            f"{UNBOUND_HEADER_PREFIX}runtime": platform.python_implementation(),
            f"{UNBOUND_HEADER_PREFIX}runtime-version": platform.python_version(),
        }

    def _build_headers(self, options: Options) -> Dict[str, Any]:
        option_headers = options.headers or {}
        headers_dict = self._merge_mappings(
            self._default_headers, option_headers, self.custom_headers
        )
        return headers_dict

    def _merge_mappings(
        self,
        *args,
    ) -> Dict[str, Any]:
        """Merge two mappings of the given type
        In cases with duplicate keys the second mapping takes precedence.
        """
        mapped_headers = {}
        for i in args:
            mapped_headers.update(i)
        return mapped_headers

    def is_closed(self) -> bool:
        return self._client.is_closed

    async def close(self) -> None:
        """Close the underlying HTTPX client.

        The client will *not* be usable after this.
        """
        if hasattr(self, "_client"):
            await self._client.aclose()

    async def __aenter__(self: Any) -> Any:
        return self

    async def __aexit__(
        self,
        exc_type: Optional[BaseException],
        exc: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        await self.close()

    async def _build_request(self, options: Options) -> httpx.Request:
        headers = self._build_headers(options)
        params = options.params
        json_body = options.json_body
        request = self._client.build_request(
            method=options.method,
            url=options.url,
            headers=headers,
            params=params,
            json=json_body,
            timeout=options.timeout,
        )
        return request

    @overload
    async def _request(
        self,
        *,
        options: Options,
        stream: Literal[False],
        cast_to: Type[ResponseT],
        stream_cls: Type[AsyncStreamT],
    ) -> ResponseT:
        ...

    @overload
    async def _request(
        self,
        *,
        options: Options,
        stream: Literal[True],
        cast_to: Type[ResponseT],
        stream_cls: Type[AsyncStreamT],
    ) -> AsyncStreamT:
        ...

    @overload
    async def _request(
        self,
        *,
        options: Options,
        stream: bool,
        cast_to: Type[ResponseT],
        stream_cls: Type[AsyncStreamT],
    ) -> Union[ResponseT, AsyncStreamT]:
        ...

    async def _request(
        self,
        *,
        options: Options,
        stream: bool,
        cast_to: Type[ResponseT],
        stream_cls: Type[AsyncStreamT],
    ) -> Union[ResponseT, AsyncStreamT]:
        request = await self._build_request(options)
        try:
            res = await self._client.send(request, auth=self.custom_auth, stream=stream)
            res.raise_for_status()
        except httpx.HTTPStatusError as err:  # 4xx and 5xx errors
            # If the response is streamed then we need to explicitly read the response
            # to completion before attempting to access the response text.
            await err.response.aread()
            raise self._make_status_error_from_response(request, err.response) from None
        except httpx.TimeoutException as err:
            raise APITimeoutError(request=request) from err
        except Exception as err:
            raise APIConnectionError(request=request) from err

        self.response_headers = res.headers
        if stream or res.headers["content-type"] == "text/event-stream":
            if stream_cls is None:
                raise MissingStreamClassError()
            stream_response = stream_cls(
                response=res, cast_to=self._extract_stream_chunk_type(stream_cls)
            )
            return stream_response

        response = (
            cast(
                ResponseT,
                cast_to(**res.json()),
            )
            if not isinstance(cast_to, httpx.Response)
            else cast(ResponseT, res)
        )
        response._headers = res.headers  # type: ignore
        return response

    def _extract_stream_chunk_type(self, stream_cls: Type) -> type:
        args = get_args(stream_cls)
        if not args:
            raise TypeError(
                f"Expected stream_cls to have been given a generic type argument, e.g. \
                    Stream[Foo] but received {stream_cls}",
            )
        return cast(type, args[0])

    def _make_status_error_from_response(
        self,
        request: httpx.Request,
        response: httpx.Response,
    ) -> APIStatusError:
        err_text = response.text.strip()
        body = err_text

        try:
            body = json.loads(err_text)["error"]["message"]
            err_msg = f"Error code: {response.status_code} - {body}"
        except Exception:
            err_msg = err_text or f"Error code: {response.status_code}"

        return make_status_error(err_msg, body=body, request=request, response=response)



from __future__ import annotations

from typing import Any
from typing_extensions import TypeGuard

import pydantic

from .._utils import is_dict as _is_dict, is_list
from .._compat import model_json_schema


def to_strict_json_schema(model: type[pydantic.BaseModel]) -> dict[str, Any]:
    return _ensure_strict_json_schema(model_json_schema(model), path=())


def _ensure_strict_json_schema(
    json_schema: object,
    path: tuple[str, ...],
) -> dict[str, Any]:
    """Mutates the given JSON schema to ensure it conforms to the `strict` standard
    that the API expects.
    """
    if not is_dict(json_schema):
        raise TypeError(f"Expected {json_schema} to be a dictionary; path={path}")

    typ = json_schema.get("type")
    if typ == "object" and "additionalProperties" not in json_schema:
        json_schema["additionalProperties"] = False

    # object types
    # { 'type': 'object', 'properties': { 'a':  {...} } }
    properties = json_schema.get("properties")
    if is_dict(properties):
        json_schema["required"] = [prop for prop in properties.keys()]
        json_schema["properties"] = {
            key: _ensure_strict_json_schema(prop_schema, path=(*path, "properties", key))
            for key, prop_schema in properties.items()
        }

    # arrays
    # { 'type': 'array', 'items': {...} }
    items = json_schema.get("items")
    if is_dict(items):
        json_schema["items"] = _ensure_strict_json_schema(items, path=(*path, "items"))

    # unions
    any_of = json_schema.get("anyOf")
    if is_list(any_of):
        json_schema["anyOf"] = [
            _ensure_strict_json_schema(variant, path=(*path, "anyOf", str(i))) for i, variant in enumerate(any_of)
        ]

    # intersections
    all_of = json_schema.get("allOf")
    if is_list(all_of):
        json_schema["allOf"] = [
            _ensure_strict_json_schema(entry, path=(*path, "anyOf", str(i))) for i, entry in enumerate(all_of)
        ]

    defs = json_schema.get("$defs")
    if is_dict(defs):
        for def_name, def_schema in defs.items():
            _ensure_strict_json_schema(def_schema, path=(*path, "$defs", def_name))

    return json_schema


def is_dict(obj: object) -> TypeGuard[dict[str, object]]:
    # just pretend that we know there are only `str` keys
    # as that check is not worth the performance cost
    return _is_dict(obj)



import json
from typing import Any, Iterable, List, Optional, Union
import typing
from unbound.api_resources.apis.api_resource import APIResource, AsyncAPIResource
from unbound.api_resources.client import AsyncUnbound, Unbound
from ..._vendor.openai._types import NotGiven, NOT_GIVEN, FileTypes
from ..._vendor.openai.types.beta import (
    vector_store_create_params,
    vector_store_update_params,
)

from unbound.api_resources.types.vector_stores_type import (
    VectorStore,
    VectorStoreDeleted,
    VectorStoreFile,
    VectorStoreFileBatch,
    VectorStoreFileDeleted,
    VectorStoreFileList,
    VectorStoreList,
)


class VectorStores(APIResource):
    def __init__(self, client: Unbound) -> None:
        super().__init__(client)
        self.openai_client = client.openai_client
        self.files = VectorFiles(client)
        self.file_batches = VectorFileBatches(client)

    def create(
        self,
        *,
        chunking_strategy: Union[Any, NotGiven] = NOT_GIVEN,
        expires_after: Union[
            vector_store_create_params.ExpiresAfter, NotGiven
        ] = NOT_GIVEN,
        file_ids: Union[List[str], NotGiven] = NOT_GIVEN,
        metadata: Union[Optional[object], NotGiven] = NOT_GIVEN,
        name: Union[str, NotGiven] = NOT_GIVEN,
    ) -> VectorStore:
        response = self.openai_client.with_raw_response.beta.vector_stores.create(
            chunking_strategy=chunking_strategy,
            expires_after=expires_after,
            file_ids=file_ids,
            metadata=metadata,
            name=name,
        )
        data = VectorStore(**json.loads(response.text))
        data._headers = response.headers

        return data

    def retrieve(
        self,
        vector_store_id: str,
        **kwargs,
    ) -> VectorStore:
        response = self.openai_client.with_raw_response.beta.vector_stores.retrieve(
            vector_store_id=vector_store_id,
            **kwargs,
        )
        data = VectorStore(**json.loads(response.text))
        data._headers = response.headers

        return data

    def update(
        self,
        vector_store_id: str,
        *,
        expires_after: Union[
            vector_store_update_params.ExpiresAfter, NotGiven
        ] = NOT_GIVEN,
        metadata: Union[Optional[object], NotGiven] = NOT_GIVEN,
        name: Union[str, NotGiven] = NOT_GIVEN,
        **kwargs,
    ) -> VectorStore:
        response = self.openai_client.with_raw_response.beta.vector_stores.update(
            vector_store_id=vector_store_id,
            expires_after=expires_after,
            metadata=metadata,
            name=name,
            **kwargs,
        )
        data = VectorStore(**json.loads(response.text))
        data._headers = response.headers

        return data

    @typing.no_type_check
    def list(
        self,
        *,
        after: Union[str, NotGiven] = NOT_GIVEN,
        limit: Union[int, NotGiven] = NOT_GIVEN,
        order: Union[str, NotGiven] = NOT_GIVEN,
        **kwargs,
    ) -> VectorStoreList:
        response = self.openai_client.with_raw_response.beta.vector_stores.list(
            after=after,
            limit=limit,
            order=order,
            **kwargs,
        )
        data = VectorStoreList(**json.loads(response.text))
        data._headers = response.headers

        return data

    def delete(
        self,
        vector_store_id: str,
        **kwargs,
    ) -> VectorStoreDeleted:
        response = self.openai_client.with_raw_response.beta.vector_stores.delete(
            vector_store_id=vector_store_id,
            **kwargs,
        )
        data = VectorStoreDeleted(**json.loads(response.text))
        data._headers = response.headers

        return data


class VectorFiles(APIResource):
    def __init__(self, client: Unbound) -> None:
        super().__init__(client)
        self.openai_client = client.openai_client

    def create(
        self,
        vector_store_id: str,
        *,
        file_id: str,
        chunking_strategy: Union[Any, NotGiven] = NOT_GIVEN,
        **kwargs,
    ) -> VectorStoreFile:
        response = self.openai_client.with_raw_response.beta.vector_stores.files.create(
            vector_store_id=vector_store_id,
            file_id=file_id,
            chunking_strategy=chunking_strategy,
            **kwargs,
        )
        data = VectorStoreFile(**json.loads(response.text))
        data._headers = response.headers

        return data

    def retrieve(
        self,
        file_id: str,
        *,
        vector_store_id: str,
        **kwargs,
    ) -> VectorStoreFile:
        response = (
            self.openai_client.with_raw_response.beta.vector_stores.files.retrieve(
                file_id=file_id,
                vector_store_id=vector_store_id,
                **kwargs,
            )
        )
        data = VectorStoreFile(**json.loads(response.text))
        data._headers = response.headers

        return data

    @typing.no_type_check
    def list(
        self,
        vector_store_id: str,
        *,
        after: Union[str, NotGiven] = NOT_GIVEN,
        before: Union[str, NotGiven] = NOT_GIVEN,
        filter: Union[str, NotGiven] = NOT_GIVEN,
        limit: Union[int, NotGiven] = NOT_GIVEN,
        order: Union[str, NotGiven] = NOT_GIVEN,
        **kwargs,
    ) -> VectorStoreFileList:
        response = self.openai_client.with_raw_response.beta.vector_stores.files.list(
            vector_store_id=vector_store_id,
            after=after,
            before=before,
            filter=filter,
            limit=limit,
            order=order,
            **kwargs,
        )
        data = VectorStoreFileList(**json.loads(response.text))
        data._headers = response.headers

        return data

    def delete(
        self,
        file_id: str,
        *,
        vector_store_id: str,
        **kwargs,
    ) -> VectorStoreFileDeleted:
        response = self.openai_client.with_raw_response.beta.vector_stores.files.delete(
            file_id=file_id,
            vector_store_id=vector_store_id,
            **kwargs,
        )
        data = VectorStoreFileDeleted(**json.loads(response.text))
        data._headers = response.headers

        return data

    def create_and_poll(
        self,
        file_id: str,
        *,
        vector_store_id: str,
        poll_interval_ms: Union[int, NotGiven] = NOT_GIVEN,
        chunking_strategy: Union[Any, NotGiven] = NOT_GIVEN,
        **kwargs,
    ) -> VectorStoreFile:
        response = self.openai_client.beta.vector_stores.files.create_and_poll(
            file_id=file_id,
            vector_store_id=vector_store_id,
            poll_interval_ms=poll_interval_ms,
            chunking_strategy=chunking_strategy,
            **kwargs,
        )
        data = response

        return data  # type: ignore[return-value]

    def poll(
        self,
        file_id: str,
        *,
        vector_store_id: str,
        poll_interval_ms: Union[int, NotGiven] = NOT_GIVEN,
        **kwargs,
    ) -> VectorStoreFile:
        response = self.openai_client.beta.vector_stores.files.poll(
            file_id=file_id,
            vector_store_id=vector_store_id,
            poll_interval_ms=poll_interval_ms,
            **kwargs,
        )
        data = response

        return data  # type: ignore[return-value]

    def upload(
        self,
        *,
        vector_store_id: str,
        file: FileTypes,
        chunking_strategy: Union[Any, NotGiven] = NOT_GIVEN,
        **kwargs,
    ) -> VectorStoreFile:
        response = self.openai_client.beta.vector_stores.files.upload(
            vector_store_id=vector_store_id,
            file=file,
            chunking_strategy=chunking_strategy,
            **kwargs,
        )
        data = response
        return data  # type: ignore[return-value]

    def upload_and_poll(
        self,
        *,
        vector_store_id: str,
        file: FileTypes,
        poll_interval_ms: Union[int, NotGiven] = NOT_GIVEN,
        chunking_strategy: Union[Any, NotGiven] = NOT_GIVEN,
        **kwargs,
    ) -> VectorStoreFile:
        response = self.openai_client.beta.vector_stores.files.upload_and_poll(
            vector_store_id=vector_store_id,
            file=file,
            poll_interval_ms=poll_interval_ms,
            chunking_strategy=chunking_strategy,
            **kwargs,
        )
        data = response
        return data  # type: ignore[return-value]


class VectorFileBatches(APIResource):
    def __init__(self, client: Unbound) -> None:
        super().__init__(client)
        self.openai_client = client.openai_client

    def create(
        self,
        vector_store_id: str,
        *,
        file_ids: List[str],
        chunking_strategy: Union[Any, NotGiven] = NOT_GIVEN,
        **kwargs,
    ) -> VectorStoreFileBatch:
        response = (
            self.openai_client.with_raw_response.beta.vector_stores.file_batches.create(
                vector_store_id=vector_store_id,
                file_ids=file_ids,
                chunking_strategy=chunking_strategy,
                **kwargs,
            )
        )
        data = VectorStoreFileBatch(**json.loads(response.text))
        data._headers = response.headers

        return data

    def retrieve(
        self, batch_id: str, *, vector_store_id: str, **kwargs
    ) -> VectorStoreFileBatch:
        response = self.openai_client.with_raw_response.beta.vector_stores.file_batches.retrieve(  # noqa: E501
            batch_id=batch_id,
            vector_store_id=vector_store_id,
            **kwargs,
        )
        data = VectorStoreFileBatch(**json.loads(response.text))
        data._headers = response.headers

        return data

    def cancel(
        self, batch_id: str, *, vector_store_id: str, **kwargs
    ) -> VectorStoreFileBatch:
        response = (
            self.openai_client.with_raw_response.beta.vector_stores.file_batches.cancel(
                batch_id=batch_id,
                vector_store_id=vector_store_id,
                **kwargs,
            )
        )
        data = VectorStoreFileBatch(**json.loads(response.text))
        data._headers = response.headers

        return data

    def create_and_poll(
        self,
        vector_store_id: str,
        *,
        file_ids: List[str],
        poll_interval_ms: Union[int, NotGiven] = NOT_GIVEN,
        chunking_strategy: Union[Any, NotGiven] = NOT_GIVEN,
        **kwargs,
    ) -> VectorStoreFileBatch:
        response = self.openai_client.beta.vector_stores.file_batches.create_and_poll(
            vector_store_id=vector_store_id,
            file_ids=file_ids,
            poll_interval_ms=poll_interval_ms,
            chunking_strategy=chunking_strategy,
            **kwargs,
        )
        data = response
        return data  # type: ignore[return-value]

    @typing.no_type_check
    def list_files(
        self,
        batch_id: str,
        *,
        vector_store_id: str,
        after: Union[str, NotGiven] = NOT_GIVEN,
        before: Union[str, NotGiven] = NOT_GIVEN,
        filter: Union[str, NotGiven] = NOT_GIVEN,
        limit: Union[int, NotGiven] = NOT_GIVEN,
        order: Union[str, NotGiven] = NOT_GIVEN,
        **kwargs,
    ) -> VectorStoreFileList:
        response = self.openai_client.with_raw_response.beta.vector_stores.file_batches.list_files(  # noqa: E501
            batch_id=batch_id,
            vector_store_id=vector_store_id,
            after=after,
            before=before,
            filter=filter,
            limit=limit,
            order=order,
            **kwargs,
        )
        data = VectorStoreFileBatch(**json.loads(response.text))
        data._headers = response.headers

        return data

    def poll(
        self,
        batch_id: str,
        *,
        vector_store_id: str,
        poll_interval_ms: Union[int, NotGiven] = NOT_GIVEN,
        **kwargs,
    ) -> VectorStoreFileBatch:
        response = self.openai_client.beta.vector_stores.file_batches.poll(
            batch_id=batch_id,
            vector_store_id=vector_store_id,
            poll_interval_ms=poll_interval_ms,
            **kwargs,
        )
        data = response

        return data  # type: ignore[return-value]

    def upload_and_poll(
        self,
        vector_store_id: str,
        *,
        files: Iterable[FileTypes],
        max_concurrency: int = 5,
        file_ids: List[str] = [],
        poll_interval_ms: Union[int, NotGiven] = NOT_GIVEN,
        chunking_strategy: Union[Any, NotGiven] = NOT_GIVEN,
        **kwargs,
    ) -> VectorStoreFileBatch:
        response = self.openai_client.beta.vector_stores.file_batches.upload_and_poll(
            vector_store_id=vector_store_id,
            files=files,
            max_concurrency=max_concurrency,
            file_ids=file_ids,
            poll_interval_ms=poll_interval_ms,
            chunking_strategy=chunking_strategy,
            **kwargs,
        )
        data = response

        return data  # type: ignore[return-value]


class AsyncVectorStores(AsyncAPIResource):
    def __init__(self, client: AsyncUnbound) -> None:
        super().__init__(client)
        self.openai_client = client.openai_client
        self.files = AsyncVectorFiles(client)
        self.file_batches = AsyncVectorFileBatches(client)

    async def create(
        self,
        *,
        chunking_strategy: Union[Any, NotGiven] = NOT_GIVEN,
        expires_after: Union[
            vector_store_create_params.ExpiresAfter, NotGiven
        ] = NOT_GIVEN,
        file_ids: Union[List[str], NotGiven] = NOT_GIVEN,
        metadata: Union[Optional[object], NotGiven] = NOT_GIVEN,
        name: Union[str, NotGiven] = NOT_GIVEN,
    ) -> VectorStore:
        response = await self.openai_client.with_raw_response.beta.vector_stores.create(
            chunking_strategy=chunking_strategy,
            expires_after=expires_after,
            file_ids=file_ids,
            metadata=metadata,
            name=name,
        )
        data = VectorStore(**json.loads(response.text))
        data._headers = response.headers

        return data

    async def retrieve(
        self,
        vector_store_id: str,
        **kwargs,
    ) -> VectorStore:
        response = (
            await self.openai_client.with_raw_response.beta.vector_stores.retrieve(
                vector_store_id=vector_store_id,
                **kwargs,
            )
        )
        data = VectorStore(**json.loads(response.text))
        data._headers = response.headers

        return data

    async def update(
        self,
        vector_store_id: str,
        *,
        expires_after: Union[
            vector_store_update_params.ExpiresAfter, NotGiven
        ] = NOT_GIVEN,
        metadata: Union[Optional[object], NotGiven] = NOT_GIVEN,
        name: Union[str, NotGiven] = NOT_GIVEN,
        **kwargs,
    ) -> VectorStore:
        response = await self.openai_client.with_raw_response.beta.vector_stores.update(
            vector_store_id=vector_store_id,
            expires_after=expires_after,
            metadata=metadata,
            name=name,
            **kwargs,
        )
        data = VectorStore(**json.loads(response.text))
        data._headers = response.headers

        return data

    @typing.no_type_check
    async def list(
        self,
        *,
        after: Union[str, NotGiven] = NOT_GIVEN,
        limit: Union[int, NotGiven] = NOT_GIVEN,
        order: Union[str, NotGiven] = NOT_GIVEN,
        **kwargs,
    ) -> VectorStoreList:
        response = await self.openai_client.with_raw_response.beta.vector_stores.list(
            after=after,
            limit=limit,
            order=order,
            **kwargs,
        )
        data = VectorStoreList(**json.loads(response.text))
        data._headers = response.headers

        return data

    async def delete(
        self,
        vector_store_id: str,
        **kwargs,
    ) -> VectorStoreDeleted:
        response = await self.openai_client.with_raw_response.beta.vector_stores.delete(
            vector_store_id=vector_store_id,
            **kwargs,
        )
        data = VectorStoreDeleted(**json.loads(response.text))
        data._headers = response.headers

        return data


class AsyncVectorFiles(AsyncAPIResource):
    def __init__(self, client: AsyncUnbound) -> None:
        super().__init__(client)
        self.openai_client = client.openai_client

    async def create(
        self,
        vector_store_id: str,
        *,
        file_id: str,
        chunking_strategy: Union[Any, NotGiven] = NOT_GIVEN,
        **kwargs,
    ) -> VectorStoreFile:
        response = (
            await self.openai_client.with_raw_response.beta.vector_stores.files.create(
                vector_store_id=vector_store_id,
                file_id=file_id,
                chunking_strategy=chunking_strategy,
                **kwargs,
            )
        )
        data = VectorStoreFile(**json.loads(response.text))
        data._headers = response.headers

        return data

    async def retrieve(
        self,
        file_id: str,
        *,
        vector_store_id: str,
        **kwargs,
    ) -> VectorStoreFile:
        response = await self.openai_client.with_raw_response.beta.vector_stores.files.retrieve(  # noqa: E501
            file_id=file_id,
            vector_store_id=vector_store_id,
            **kwargs,
        )
        data = VectorStoreFile(**json.loads(response.text))
        data._headers = response.headers

        return data

    @typing.no_type_check
    async def list(
        self,
        vector_store_id: str,
        *,
        after: Union[str, NotGiven] = NOT_GIVEN,
        before: Union[str, NotGiven] = NOT_GIVEN,
        filter: Union[str, NotGiven] = NOT_GIVEN,
        limit: Union[int, NotGiven] = NOT_GIVEN,
        order: Union[str, NotGiven] = NOT_GIVEN,
        **kwargs,
    ) -> VectorStoreFileList:
        response = (
            await self.openai_client.with_raw_response.beta.vector_stores.files.list(
                vector_store_id=vector_store_id,
                after=after,
                before=before,
                filter=filter,
                limit=limit,
                order=order,
                **kwargs,
            )
        )
        data = VectorStoreFileList(**json.loads(response.text))
        data._headers = response.headers

        return data

    async def delete(
        self,
        file_id: str,
        *,
        vector_store_id: str,
        **kwargs,
    ) -> VectorStoreFileDeleted:
        response = (
            await self.openai_client.with_raw_response.beta.vector_stores.files.delete(
                file_id=file_id,
                vector_store_id=vector_store_id,
                **kwargs,
            )
        )
        data = VectorStoreFileDeleted(**json.loads(response.text))
        data._headers = response.headers

        return data

    async def create_and_poll(
        self,
        file_id: str,
        *,
        vector_store_id: str,
        poll_interval_ms: Union[int, NotGiven] = NOT_GIVEN,
        chunking_strategy: Union[Any, NotGiven] = NOT_GIVEN,
        **kwargs,
    ) -> VectorStoreFile:
        response = await self.openai_client.beta.vector_stores.files.create_and_poll(
            file_id=file_id,
            vector_store_id=vector_store_id,
            poll_interval_ms=poll_interval_ms,
            chunking_strategy=chunking_strategy,
            **kwargs,
        )
        data = response

        return data  # type: ignore[return-value]

    async def poll(
        self,
        file_id: str,
        *,
        vector_store_id: str,
        poll_interval_ms: Union[int, NotGiven] = NOT_GIVEN,
        **kwargs,
    ) -> VectorStoreFile:
        response = await self.openai_client.beta.vector_stores.files.poll(
            file_id=file_id,
            vector_store_id=vector_store_id,
            poll_interval_ms=poll_interval_ms,
            **kwargs,
        )
        data = response

        return data  # type: ignore[return-value]

    async def upload(
        self,
        *,
        vector_store_id: str,
        file: FileTypes,
        chunking_strategy: Union[Any, NotGiven] = NOT_GIVEN,
        **kwargs,
    ) -> VectorStoreFile:
        response = await self.openai_client.beta.vector_stores.files.upload(
            vector_store_id=vector_store_id,
            file=file,
            chunking_strategy=chunking_strategy,
            **kwargs,
        )
        data = response
        return data  # type: ignore[return-value]

    async def upload_and_poll(
        self,
        *,
        vector_store_id: str,
        file: FileTypes,
        poll_interval_ms: Union[int, NotGiven] = NOT_GIVEN,
        chunking_strategy: Union[Any, NotGiven] = NOT_GIVEN,
        **kwargs,
    ) -> VectorStoreFile:
        response = await self.openai_client.beta.vector_stores.files.upload_and_poll(
            vector_store_id=vector_store_id,
            file=file,
            poll_interval_ms=poll_interval_ms,
            chunking_strategy=chunking_strategy,
            **kwargs,
        )
        data = response
        return data  # type: ignore[return-value]


class AsyncVectorFileBatches(AsyncAPIResource):
    def __init__(self, client: AsyncUnbound) -> None:
        super().__init__(client)
        self.openai_client = client.openai_client

    async def create(
        self,
        vector_store_id: str,
        *,
        file_ids: List[str],
        chunking_strategy: Union[Any, NotGiven] = NOT_GIVEN,
        **kwargs,
    ) -> VectorStoreFileBatch:
        response = await self.openai_client.with_raw_response.beta.vector_stores.file_batches.create(  # noqa: E501
            vector_store_id=vector_store_id,
            file_ids=file_ids,
            chunking_strategy=chunking_strategy,
            **kwargs,
        )
        data = VectorStoreFileBatch(**json.loads(response.text))
        data._headers = response.headers

        return data

    async def retrieve(
        self, batch_id: str, *, vector_store_id: str, **kwargs
    ) -> VectorStoreFileBatch:
        response = await self.openai_client.with_raw_response.beta.vector_stores.file_batches.retrieve(  # noqa: E501
            batch_id=batch_id,
            vector_store_id=vector_store_id,
            **kwargs,
        )
        data = VectorStoreFileBatch(**json.loads(response.text))
        data._headers = response.headers

        return data

    async def cancel(
        self, batch_id: str, *, vector_store_id: str, **kwargs
    ) -> VectorStoreFileBatch:
        response = await self.openai_client.with_raw_response.beta.vector_stores.file_batches.cancel(  # noqa: E501
            batch_id=batch_id,
            vector_store_id=vector_store_id,
            **kwargs,
        )
        data = VectorStoreFileBatch(**json.loads(response.text))
        data._headers = response.headers

        return data

    async def create_and_poll(
        self,
        vector_store_id: str,
        *,
        file_ids: List[str],
        poll_interval_ms: Union[int, NotGiven] = NOT_GIVEN,
        chunking_strategy: Union[Any, NotGiven] = NOT_GIVEN,
        **kwargs,
    ) -> VectorStoreFileBatch:
        response = (
            await self.openai_client.beta.vector_stores.file_batches.create_and_poll(
                vector_store_id=vector_store_id,
                file_ids=file_ids,
                poll_interval_ms=poll_interval_ms,
                chunking_strategy=chunking_strategy,
                **kwargs,
            )
        )
        data = response

        return data  # type: ignore[return-value]

    @typing.no_type_check
    async def list_files(
        self,
        batch_id: str,
        *,
        vector_store_id: str,
        after: Union[str, NotGiven] = NOT_GIVEN,
        before: Union[str, NotGiven] = NOT_GIVEN,
        filter: Union[str, NotGiven] = NOT_GIVEN,
        limit: Union[int, NotGiven] = NOT_GIVEN,
        order: Union[str, NotGiven] = NOT_GIVEN,
        **kwargs,
    ) -> VectorStoreFileList:
        response = await self.openai_client.beta.with_raw_response.vector_stores.file_batches.list_files(  # noqa: E501
            batch_id=batch_id,
            vector_store_id=vector_store_id,
            after=after,
            before=before,
            filter=filter,
            limit=limit,
            order=order,
            **kwargs,
        )
        data = VectorStoreFileBatch(**json.loads(response.text))
        data._headers = response.headers

        return data

    async def poll(
        self,
        batch_id: str,
        *,
        vector_store_id: str,
        poll_interval_ms: Union[int, NotGiven] = NOT_GIVEN,
        **kwargs,
    ) -> VectorStoreFileBatch:
        response = await self.openai_client.beta.vector_stores.file_batches.poll(
            batch_id=batch_id,
            vector_store_id=vector_store_id,
            poll_interval_ms=poll_interval_ms,
            **kwargs,
        )
        data = response

        return data  # type: ignore[return-value]

    async def upload_and_poll(
        self,
        vector_store_id: str,
        *,
        files: Iterable[FileTypes],
        max_concurrency: int = 5,
        file_ids: List[str] = [],
        poll_interval_ms: Union[int, NotGiven] = NOT_GIVEN,
        chunking_strategy: Union[Any, NotGiven] = NOT_GIVEN,
        **kwargs,
    ) -> VectorStoreFileBatch:
        response = (
            await self.openai_client.beta.vector_stores.file_batches.upload_and_poll(
                vector_store_id=vector_store_id,
                files=files,
                max_concurrency=max_concurrency,
                file_ids=file_ids,
                poll_interval_ms=poll_interval_ms,
                chunking_strategy=chunking_strategy,
                **kwargs,
            )
        )
        data = response

        return data  # type: ignore[return-value]



from __future__ import annotations

from typing import Optional
from unbound.api_resources import apis
from unbound.api_resources.base_client import APIClient, AsyncAPIClient

# from openai import AsyncOpenAI, OpenAI
from .._vendor.openai import OpenAI, AsyncOpenAI


class Unbound(APIClient):
    completions: apis.Completion
    chat: apis.ChatCompletion
    generations: apis.Generations
    prompts: apis.Prompts
    embeddings: apis.Embeddings
    feedback: apis.Feedback
    images: apis.Images
    files: apis.MainFiles
    models: apis.Models
    moderations: apis.Moderations
    audio: apis.Audio
    batches: apis.Batches
    fine_tuning: apis.FineTuning
    admin: apis.Admin
    uploads: apis.Uploads
    configs: apis.Configs
    api_keys: apis.ApiKeys
    virtual_keys: apis.VirtualKeys
    logs: apis.Logs

    class beta:
        assistants: apis.Assistants
        threads: apis.Threads
        vector_stores: apis.VectorStores
        chat: apis.BetaChat

        def __init__(self, client: Unbound) -> None:
            self.assistants = apis.Assistants(client)
            self.threads = apis.Threads(client)
            self.vector_stores = apis.VectorStores(client)
            self.chat = apis.BetaChat(client)

    def __init__(
        self,
        *,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        **kwargs,
    ) -> None:
        super().__init__(
            api_key=api_key,
            base_url=base_url,
            **kwargs,
        )

        self.openai_client = OpenAI(
            api_key=api_key,
            base_url=self.base_url,
            default_headers=self.allHeaders,
            max_retries=0,
        )

        self.completions = apis.Completion(self)
        self.chat = apis.ChatCompletion(self)
        self.generations = apis.Generations(self)
        self.prompts = apis.Prompts(self)
        self.embeddings = apis.Embeddings(self)
        self.feedback = apis.Feedback(self)
        self.images = apis.Images(self)
        self.files = apis.MainFiles(self)
        self.models = apis.Models(self)
        self.moderations = apis.Moderations(self)
        self.audio = apis.Audio(self)
        self.batches = apis.Batches(self)
        self.fine_tuning = apis.FineTuning(self)
        self.admin = apis.Admin(self)
        self.uploads = apis.Uploads(self)
        self.configs = apis.Configs(self)
        self.api_keys = apis.ApiKeys(self)
        self.virtual_keys = apis.VirtualKeys(self)
        self.logs = apis.Logs(self)
        self.beta = self.beta(self)  # type: ignore

    def copy(
        self,
        *,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        **kwargs,
    ) -> Unbound:
        return self.__class__(
            api_key=api_key or self.api_key,
            base_url=base_url or self.base_url,
            **kwargs,
        )

    def post(self, url: str, **kwargs):
        return apis.Post(self).create(url=url, **kwargs)

    with_options = copy


class AsyncUnbound(AsyncAPIClient):
    completions: apis.AsyncCompletion
    chat: apis.AsyncChatCompletion
    generations: apis.AsyncGenerations
    prompts: apis.AsyncPrompts
    embeddings: apis.AsyncEmbeddings
    feedback: apis.AsyncFeedback
    images: apis.AsyncImages
    files: apis.AsyncMainFiles
    models: apis.AsyncModels
    moderations: apis.AsyncModerations
    audio: apis.AsyncAudio
    batches: apis.AsyncBatches
    fine_tuning: apis.AsyncFineTuning
    admin: apis.AsyncAdmin
    uploads: apis.AsyncUploads
    configs: apis.AsyncConfigs
    api_keys: apis.AsyncApiKeys
    virtual_keys: apis.AsyncVirtualKeys
    logs: apis.AsyncLogs

    class beta:
        assistants: apis.AsyncAssistants
        threads: apis.AsyncThreads
        vector_stores: apis.AsyncVectorStores
        chat: apis.AsyncBetaChat

        def __init__(self, client: AsyncUnbound) -> None:
            self.assistants = apis.AsyncAssistants(client)
            self.threads = apis.AsyncThreads(client)
            self.vector_stores = apis.AsyncVectorStores(client)
            self.chat = apis.AsyncBetaChat(client)

    def __init__(
        self,
        *,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        **kwargs,
    ) -> None:
        super().__init__(
            api_key=api_key,
            base_url=base_url,
            **kwargs,
        )

        self.openai_client = AsyncOpenAI(
            api_key=api_key,
            base_url=self.base_url,
            default_headers=self.allHeaders,
            max_retries=0,
        )

        self.completions = apis.AsyncCompletion(self)
        self.chat = apis.AsyncChatCompletion(self)
        self.generations = apis.AsyncGenerations(self)
        self.prompts = apis.AsyncPrompts(self)
        self.embeddings = apis.AsyncEmbeddings(self)
        self.feedback = apis.AsyncFeedback(self)
        self.images = apis.AsyncImages(self)
        self.files = apis.AsyncMainFiles(self)
        self.models = apis.AsyncModels(self)
        self.moderations = apis.AsyncModerations(self)
        self.audio = apis.AsyncAudio(self)
        self.batches = apis.AsyncBatches(self)
        self.fine_tuning = apis.AsyncFineTuning(self)
        self.admin = apis.AsyncAdmin(self)
        self.uploads = apis.AsyncUploads(self)
        self.configs = apis.AsyncConfigs(self)
        self.api_keys = apis.AsyncApiKeys(self)
        self.virtual_keys = apis.AsyncVirtualKeys(self)
        self.logs = apis.AsyncLogs(self)
        self.beta = self.beta(self)  # type: ignore

    def copy(
        self,
        *,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        **kwargs,
    ) -> AsyncUnbound:
        return self.__class__(
            api_key=api_key or self.api_key,
            base_url=base_url or self.base_url,
            **kwargs,
        )

    async def post(self, url: str, **kwargs):
        return await apis.AsyncPost(self).create(url=url, **kwargs)

    with_options = copy



from typing import Any, Dict, Optional
from unbound.api_resources.base_client import APIClient, AsyncAPIClient
from urllib.parse import urlencode
from unbound.api_resources.apis.api_resource import APIResource, AsyncAPIResource
from unbound.api_resources.types.configs_type import (
    ConfigAddResponse,
    ConfigGetResponse,
    ConfigListResponse,
    ConfigUpdateResponse,
)
from unbound.api_resources.utils import GenericResponse
from unbound.api_resources.utils import UnboundApiPaths


class Configs(APIResource):
    def __init__(self, client: APIClient) -> None:
        super().__init__(client)

    def create(
        self,
        *,
        name: Optional[str] = None,
        config: Optional[Dict[str, Any]] = None,
        is_default: Optional[int] = None,
        workspace_id: Optional[str] = None,
    ) -> ConfigAddResponse:
        body = {
            "name": name,
            "config": config,
            "is_default": is_default,
            "workspace_id": workspace_id,
        }
        return self._post(
            f"{UnboundApiPaths.CONFIG_API}",
            body=body,
            params=None,
            cast_to=ConfigAddResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    def retrieve(self, *, slug: Optional[str]) -> ConfigGetResponse:
        return self._get(
            f"{UnboundApiPaths.CONFIG_API}/{slug}",
            params=None,
            body=None,
            cast_to=ConfigGetResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    def list(
        self,
        *,
        workspace_id: Optional[str] = None,
    ) -> ConfigListResponse:
        query = {
            "workspace_id": workspace_id,
        }
        filtered_query = {k: v for k, v in query.items() if v is not None}
        query_string = urlencode(filtered_query)
        return self._get(
            f"{UnboundApiPaths.CONFIG_API}?{query_string}",
            params=None,
            body=None,
            cast_to=ConfigListResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    def update(
        self,
        *,
        slug: Optional[str] = None,
        name: Optional[str] = None,
        config: Optional[Dict[str, Any]] = None,
        status: Optional[str] = None,
    ) -> ConfigUpdateResponse:
        body = {
            "slug": slug,
            "name": name,
            "config": config,
            "status": status,
        }
        return self._put(
            f"{UnboundApiPaths.CONFIG_API}/{slug}",
            body=body,
            params=None,
            cast_to=ConfigUpdateResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    def delete(
        self,
        *,
        id: Optional[str],
    ) -> Any:
        return self._delete(
            f"{UnboundApiPaths.CONFIG_API}/{id}",
            params=None,
            body=None,
            cast_to=GenericResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )


class AsyncConfigs(AsyncAPIResource):
    def __init__(self, client: AsyncAPIClient) -> None:
        super().__init__(client)

    async def create(
        self,
        *,
        name: Optional[str] = None,
        config: Optional[Dict[str, Any]] = None,
        is_default: Optional[int] = None,
        workspace_id: Optional[str] = None,
    ) -> ConfigAddResponse:
        body = {
            "name": name,
            "config": config,
            "is_default": is_default,
            "workspace_id": workspace_id,
        }
        return await self._post(
            f"{UnboundApiPaths.CONFIG_API}",
            body=body,
            params=None,
            cast_to=ConfigAddResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    async def retrieve(self, *, slug: Optional[str]) -> ConfigGetResponse:
        return await self._get(
            f"{UnboundApiPaths.CONFIG_API}/{slug}",
            params=None,
            body=None,
            cast_to=ConfigGetResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    async def list(
        self,
        *,
        workspace_id: Optional[str] = None,
    ) -> ConfigListResponse:
        query = {
            "workspace_id": workspace_id,
        }
        filtered_query = {k: v for k, v in query.items() if v is not None}
        query_string = urlencode(filtered_query)
        return await self._get(
            f"{UnboundApiPaths.CONFIG_API}?{query_string}",
            params=None,
            body=None,
            cast_to=ConfigListResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    async def update(
        self,
        *,
        slug: Optional[str] = None,
        name: Optional[str] = None,
        config: Optional[Dict[str, Any]] = None,
        status: Optional[str] = None,
    ) -> ConfigUpdateResponse:
        body = {
            "slug": slug,
            "name": name,
            "config": config,
            "status": status,
        }
        return await self._put(
            f"{UnboundApiPaths.CONFIG_API}/{slug}",
            body=body,
            params=None,
            cast_to=ConfigUpdateResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    async def delete(
        self,
        *,
        id: Optional[str],
    ) -> Any:
        return await self._delete(
            f"{UnboundApiPaths.CONFIG_API}/{id}",
            params=None,
            body=None,
            cast_to=GenericResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )



import json
from typing import Dict, List, Optional, Union
import httpx
from .utils import parse_headers
from pydantic import BaseModel, PrivateAttr

__all__ = [
    "ThreadMessage",
    "MessageList",
    "Content",
    "Text",
    "TextAnnotation",
    "TextAnnotationFileCitation",
    "TextAnnotationFileCitationFileCitation",
    "TextAnnotationFilePath",
    "TextAnnotationFilePathFilePath",
    "MessageContentImageFile",
    "ImageFile",
    "MessageContentText",
]


class TextAnnotationFilePathFilePath(BaseModel):
    file_id: Optional[str]


class TextAnnotationFileCitationFileCitation(BaseModel):
    file_id: Optional[str]
    quote: Optional[str]


class TextAnnotationFilePath(BaseModel):
    end_index: Optional[int]
    file_path: Optional[TextAnnotationFilePathFilePath]
    start_index: Optional[int]
    text: Optional[str]
    type: Optional[str]


class TextAnnotationFileCitation(BaseModel):
    end_index: Optional[int]
    file_citation: Optional[TextAnnotationFileCitationFileCitation]
    start_index: Optional[int]
    text: Optional[str]
    type: Optional[str]


TextAnnotation = Union[TextAnnotationFileCitation, TextAnnotationFilePath]


class Text(BaseModel):
    annotations: Optional[List[TextAnnotation]]
    value: Optional[str]


class MessageContentText(BaseModel):
    text: Optional[Text]
    type: Optional[str]


class ImageFile(BaseModel):
    file_id: Optional[str]


class MessageContentImageFile(BaseModel):
    image_file: Optional[ImageFile]
    type: Optional[str]


Content = Union[MessageContentImageFile, MessageContentText]


class ThreadMessage(BaseModel, extra="allow"):
    id: Optional[str]
    assistant_id: Optional[str] = None
    content: Optional[List[Content]]
    created_at: Optional[int]
    file_ids: Optional[List[str]] = None
    metadata: Optional[object] = None
    object: Optional[str]
    role: Optional[str]
    run_id: Optional[str] = None
    thread_id: Optional[str]
    _headers: Optional[httpx.Headers] = PrivateAttr()

    def __str__(self):
        del self._headers
        return json.dumps(self.dict(), indent=4)

    def get_headers(self) -> Optional[Dict[str, str]]:
        return parse_headers(self._headers)


class MessageList(BaseModel, extra="allow"):
    object: Optional[str]
    data: Optional[List[ThreadMessage]]
    first_id: Optional[str]
    last_id: Optional[str]
    has_more: Optional[bool]
    _headers: Optional[httpx.Headers] = PrivateAttr()

    def __str__(self):
        del self._headers
        return json.dumps(self.dict(), indent=4)

    def get_headers(self) -> Optional[Dict[str, str]]:
        return parse_headers(self._headers)


class ThreadMessageDeleted(BaseModel):
    id: str
    deleted: bool
    object: str
    _headers: Optional[httpx.Headers] = PrivateAttr()

    def __str__(self):
        del self._headers
        return json.dumps(self.dict(), indent=4)

    def get_headers(self) -> Optional[Dict[str, str]]:
        return parse_headers(self._headers)



from __future__ import annotations
from unbound import Unbound

import logging
from typing import Any, Iterator, List, Mapping, Optional, Union

try:
    from langchain.callbacks.manager import CallbackManagerForLLMRun
    from langchain.llms.base import LLM
    from langchain.pydantic_v1 import Field, PrivateAttr
    from langchain.schema.output import GenerationChunk
except ImportError as exc:
    raise Exception(
        "Langchain is not installed.Please install it with `pip install langchain`."
    ) from exc


logger = logging.getLogger(__name__)


class UnboundLLM(LLM):
    """Unbound Service models
    To use, you should have the ``unbound`` python package installed, and the
    environment variable ``UNBOUND_API_KEY``, set with your API key, or pass
    it as a named parameter to the `Unbound` constructor.
    NOTE: You can install unbound using ``pip install unbound``
    Example:
        .. code-block:: python
            import unbound
            from langchain.llms import Unbound
            # Simplest invocation for an openai provider. Can be extended to
            # others as well
            llm_option = unbound.LLMOptions(
                provider="openai",
                # Checkout the docs for the virtual-api-key
                virtual_key="openai-virtual-key",
                model="text-davinci-003"
            )
            # Initialise the client
            client = Unbound(
                api_key="UNBOUND_API_KEY",
                mode="single"
            ).add_llms(llm_params=llm_option)
            response = client("What are the biggest risks facing humanity?")
    """

    model: Optional[str] = Field(default="gpt-3.5-turbo")
    _client: Any = PrivateAttr()

    api_key: Optional[str] = None
    base_url: Optional[str] = None
    virtual_key: Optional[str] = None
    config: Optional[Union[Mapping, str]] = None
    provider: Optional[str] = None
    trace_id: Optional[str] = None
    custom_metadata: Optional[str] = None
    streaming: bool = False

    def __init__(
        self,
        *,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        virtual_key: Optional[str] = None,
        config: Optional[Union[Mapping, str]] = None,
        provider: Optional[str] = None,
        trace_id: Optional[str] = None,
        custom_metadata: Optional[str] = None,
        **kwargs,
    ) -> None:
        super().__init__()

        self._client = Unbound(
            api_key=api_key,
            base_url=base_url,
            virtual_key=virtual_key,
            config=config,
            provider=provider,
            trace_id=trace_id,
            metadata=custom_metadata,
            **kwargs,
        )
        self.model = None

    def _call(
        self,
        prompt: str,
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> str:
        """Call Unbound's completions endpoint.
        Args:
            prompt: The prompt to pass into the model.
            stop: Optional list of stop words to use when generating.
        Returns:
            The string generated by the provider set in the initialisation of the LLM.
        Example:
            .. code-block:: python
                response = unbound("Tell me a joke.")
        """
        response = self._client.completions.create(
            prompt=prompt, stream=False, stop=stop, **kwargs
        )
        text = response.choices[0].text
        return text or ""

    def _stream(
        self,
        prompt: str,
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> Iterator[GenerationChunk]:
        """Call Unbound completion_stream and return the resulting generator.
        Args:
            prompt: The prompt to pass into the model.
            stop: Optional list of stop words to use when generating.
        Returns:
            A generator representing the stream of tokens from Unbound.
        Example:
            .. code-block:: python
                prompt = "Write a poem about a stream."
                generator = unbound.stream(prompt)
                for token in generator:
                    yield token
        """
        response = self._client.completions.create(
            stream=True, prompt=prompt, stop=stop, **kwargs
        )
        for token in response:
            chunk = GenerationChunk(text=token.choices[0].text or "")
            yield chunk
            if run_manager:
                run_manager.on_llm_new_token(chunk.text, chunk=chunk)

    @property
    def _llm_type(self) -> str:
        """Return type of llm."""
        return "unbound-gateway"



MISSING_API_KEY_ERROR_MESSAGE = """Unbound API Key Not Found \

Resolution: \

1. Get your Unbound API key from https://gateway.unboundsecurity.ai/api-keys
2. Pass it while instantiating the Unbound client with api_key param,\
 or set it as an environment variable with export UNBOUND_API_KEY=YOUR_API_KEY
"""

MISSING_BASE_URL = """No Base url provided. Please provide a valid base url.
For example: https://gateway.unboundsecurity.ai
"""

MISSING_CONFIG_MESSAGE = (
    """The 'config' parameter is not set. Please provide a valid Config object."""
)
MISSING_MODE_MESSAGE = (
    """The 'mode' parameter is not set. Please provide a valid mode literal."""
)

INVALID_UNBOUND_MODE = """
Argument of type '{}' cannot be assigned to parameter "mode" of \
    type "ModesLiteral | Modes | None"
"""

LOCALHOST_CONNECTION_ERROR = """Could not instantiate the Unbound client. \
You can either add a valid `api_key` parameter (from https://gateway.unboundsecurity.ai/api-keys)\
or check the `base_url` parameter in the Unbound client, \
for your AI Gateway's instance's URL.
"""

CUSTOM_HOST_CONNECTION_ERROR = """We could not connect to the AI Gateway's instance. \
Please check the `base_url` parameter in the Unbound client.
"""

DEFAULT_MAX_RETRIES = 2
VERSION = "0.1.0"
DEFAULT_TIMEOUT = 60
UNBOUND_HEADER_PREFIX = "x-unbound-"
UNBOUND_BASE_URL = "https://gateway.unboundsecurity.ai/v1"
UNBOUND_GATEWAY_URL = UNBOUND_BASE_URL
LOCAL_BASE_URL = "http://localhost:8787/v1"
UNBOUND_API_KEY_ENV = "UNBOUND_API_KEY"
UNBOUND_PROXY_ENV = "UNBOUND_PROXY"
OPEN_AI_API_KEY = "OPENAI_API_KEY"



from typing import Any, Dict, List, Optional
from unbound.api_resources.base_client import APIClient, AsyncAPIClient
from urllib.parse import urlencode
from unbound.api_resources.apis.api_resource import APIResource, AsyncAPIResource
from unbound.api_resources.types.logs_type import (
    LogsExportCreateResponse,
    LogsExportListResponse,
    LogsExportUpdateResponse,
    LogsExportCancelResponse,
    LogsExportRetrieveResponse,
    LogsExportStartResponse,
    LogsExportDownloadResponse,
)
from unbound.api_resources.utils import UnboundApiPaths


class Logs(APIResource):
    def __init__(self, client: APIClient) -> None:
        super().__init__(client)
        self.exports = Exports(client)


class Exports(APIResource):
    def __init__(self, client: APIClient) -> None:
        super().__init__(client)

    def create(
        self,
        *,
        filters: Optional[Dict[str, Any]] = None,
        workspace_id: Optional[str] = None,
        description: Optional[str] = None,
        requested_data: Optional[List[str]] = None,
    ) -> LogsExportCreateResponse:
        body = {
            "filters": filters,
            "workspace_id": workspace_id,
            "description": description,
            "requested_data": requested_data,
        }
        return self._post(
            f"{UnboundApiPaths.LOGS_EXPORT_API}",
            body=body,
            params=None,
            cast_to=LogsExportCreateResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    def retrieve(self, *, export_id: str) -> LogsExportRetrieveResponse:
        return self._get(
            f"{UnboundApiPaths.LOGS_EXPORT_API}/{export_id}",
            params=None,
            body=None,
            cast_to=LogsExportRetrieveResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    def list(
        self,
        *,
        workspace_id: Optional[str] = None,
    ) -> LogsExportListResponse:
        query = {
            "workspace_id": workspace_id,
        }
        filtered_query = {k: v for k, v in query.items() if v is not None}
        query_string = urlencode(filtered_query)
        return self._get(
            f"{UnboundApiPaths.LOGS_EXPORT_API}?{query_string}",
            params=None,
            body=None,
            cast_to=LogsExportListResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    def update(
        self,
        *,
        export_id: Optional[str] = None,
        workspace_id: Optional[str] = None,
        filters: Optional[Dict[str, Any]] = None,
        requested_data: Optional[List[str]] = None,
    ) -> LogsExportUpdateResponse:
        body = {
            "workspace_id": workspace_id,
            "filters": filters,
            "requested_data": requested_data,
        }
        return self._put(
            f"{UnboundApiPaths.LOGS_EXPORT_API}/{export_id}",
            body=body,
            params=None,
            cast_to=LogsExportUpdateResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    def start(
        self,
        *,
        export_id: Optional[str],
    ) -> LogsExportStartResponse:
        return self._post(
            f"{UnboundApiPaths.LOGS_EXPORT_API}/{export_id}/start",
            body=None,
            params=None,
            cast_to=LogsExportStartResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    def cancel(
        self,
        *,
        export_id: Optional[str],
    ) -> LogsExportCancelResponse:
        return self._post(
            f"{UnboundApiPaths.LOGS_EXPORT_API}/{export_id}/cancel",
            body=None,
            params=None,
            cast_to=LogsExportCancelResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    def download(
        self,
        *,
        export_id: Optional[str],
    ) -> LogsExportDownloadResponse:
        return self._get(
            f"{UnboundApiPaths.LOGS_EXPORT_API}/{export_id}/download",
            params=None,
            body=None,
            cast_to=LogsExportDownloadResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )


class AsyncLogs(AsyncAPIResource):
    def __init__(self, client: AsyncAPIClient) -> None:
        super().__init__(client)
        self.exports = AsyncExports(client)


class AsyncExports(AsyncAPIResource):
    def __init__(self, client: AsyncAPIClient) -> None:
        super().__init__(client)

    async def create(
        self,
        *,
        filters: Optional[Dict[str, Any]] = None,
        workspace_id: Optional[str] = None,
        description: Optional[str] = None,
        requested_data: Optional[List[str]] = None,
    ) -> LogsExportCreateResponse:
        body = {
            "filters": filters,
            "workspace_id": workspace_id,
            "description": description,
            "requested_data": requested_data,
        }
        return await self._post(
            f"{UnboundApiPaths.LOGS_EXPORT_API}",
            body=body,
            params=None,
            cast_to=LogsExportCreateResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    async def retrieve(self, *, export_id: str) -> LogsExportRetrieveResponse:
        return await self._get(
            f"{UnboundApiPaths.LOGS_EXPORT_API}/{export_id}",
            params=None,
            body=None,
            cast_to=LogsExportRetrieveResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    async def list(
        self,
        *,
        workspace_id: Optional[str] = None,
    ) -> LogsExportListResponse:
        query = {
            "workspace_id": workspace_id,
        }
        filtered_query = {k: v for k, v in query.items() if v is not None}
        query_string = urlencode(filtered_query)
        return await self._get(
            f"{UnboundApiPaths.LOGS_EXPORT_API}?{query_string}",
            params=None,
            body=None,
            cast_to=LogsExportListResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    async def update(
        self,
        *,
        export_id: Optional[str] = None,
        workspace_id: Optional[str] = None,
        filters: Optional[Dict[str, Any]] = None,
        requested_data: Optional[List[str]] = None,
    ) -> LogsExportUpdateResponse:
        body = {
            "workspace_id": workspace_id,
            "filters": filters,
            "requested_data": requested_data,
        }
        return await self._put(
            f"{UnboundApiPaths.LOGS_EXPORT_API}/{export_id}",
            body=body,
            params=None,
            cast_to=LogsExportUpdateResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    async def start(
        self,
        *,
        export_id: Optional[str],
    ) -> LogsExportStartResponse:
        return await self._post(
            f"{UnboundApiPaths.LOGS_EXPORT_API}/{export_id}/start",
            body=None,
            params=None,
            cast_to=LogsExportStartResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    async def cancel(
        self,
        *,
        export_id: Optional[str],
    ) -> LogsExportCancelResponse:
        return await self._post(
            f"{UnboundApiPaths.LOGS_EXPORT_API}/{export_id}/cancel",
            body=None,
            params=None,
            cast_to=LogsExportCancelResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    async def download(
        self,
        *,
        export_id: Optional[str],
    ) -> LogsExportDownloadResponse:
        return await self._get(
            f"{UnboundApiPaths.LOGS_EXPORT_API}/{export_id}/download",
            params=None,
            body=None,
            cast_to=LogsExportDownloadResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing_extensions import Literal, TypeAlias

__all__ = ["ModerationModel"]

ModerationModel: TypeAlias = Literal["text-moderation-latest", "text-moderation-stable"]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Union
from typing_extensions import TypeAlias

from .chat_completion_tool_message_param import ChatCompletionToolMessageParam
from .chat_completion_user_message_param import ChatCompletionUserMessageParam
from .chat_completion_system_message_param import ChatCompletionSystemMessageParam
from .chat_completion_function_message_param import ChatCompletionFunctionMessageParam
from .chat_completion_assistant_message_param import ChatCompletionAssistantMessageParam

__all__ = ["ChatCompletionMessageParam"]

ChatCompletionMessageParam: TypeAlias = Union[
    ChatCompletionSystemMessageParam,
    ChatCompletionUserMessageParam,
    ChatCompletionAssistantMessageParam,
    ChatCompletionToolMessageParam,
    ChatCompletionFunctionMessageParam,
]



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing import Dict, Optional
from typing_extensions import Literal, Required, TypedDict

__all__ = ["BatchCreateParams"]


class BatchCreateParams(TypedDict, total=False):
    completion_window: Required[Literal["24h"]]
    """The time frame within which the batch should be processed.

    Currently only `24h` is supported.
    """

    endpoint: Required[Literal["/v1/chat/completions", "/v1/embeddings", "/v1/completions"]]
    """The endpoint to be used for all requests in the batch.

    Currently `/v1/chat/completions`, `/v1/embeddings`, and `/v1/completions` are
    supported. Note that `/v1/embeddings` batches are also restricted to a maximum
    of 50,000 embedding inputs across all requests in the batch.
    """

    input_file_id: Required[str]
    """The ID of an uploaded file that contains requests for the new batch.

    See [upload file](https://platform.openai.com/docs/api-reference/files/create)
    for how to upload a file.

    Your input file must be formatted as a
    [JSONL file](https://platform.openai.com/docs/api-reference/batch/request-input),
    and must be uploaded with the purpose `batch`. The file can contain up to 50,000
    requests, and can be up to 100 MB in size.
    """

    metadata: Optional[Dict[str, str]]
    """Optional custom metadata for the batch."""



import json
from typing import Dict, Optional, Union
import httpx

from .utils import parse_headers
from typing import List, Any
from pydantic import BaseModel, PrivateAttr

__all__ = ["CreateEmbeddingResponse", "Usage", "Embedding"]


class Usage(BaseModel, extra="allow"):
    prompt_tokens: Optional[int] = None
    completion_tokens: Optional[int] = None
    total_tokens: Optional[int] = None


class Embedding(BaseModel):
    embedding: Union[List[float], str]
    index: Optional[int]
    object: Optional[str]


class CreateEmbeddingResponse(BaseModel):
    success: Optional[bool] = None
    warning: Optional[str] = None
    data: List[Embedding]
    model: Optional[str]
    object: Optional[str]
    usage: Usage
    _headers: Optional[httpx.Headers] = PrivateAttr()

    def __str__(self):
        del self._headers
        return json.dumps(self.dict(), indent=4)

    def __getitem__(self, key):
        return getattr(self, key, None)

    def get(self, key: str, default: Optional[Any] = None):
        return getattr(self, key, None) or default

    def get_headers(self) -> Optional[Dict[str, str]]:
        return parse_headers(self._headers)



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from .batch import Batch as Batch
from .image import Image as Image
from .model import Model as Model
from .shared import (
    ErrorObject as ErrorObject,
    FunctionDefinition as FunctionDefinition,
    FunctionParameters as FunctionParameters,
    ResponseFormatText as ResponseFormatText,
    ResponseFormatJSONObject as ResponseFormatJSONObject,
    ResponseFormatJSONSchema as ResponseFormatJSONSchema,
)
from .upload import Upload as Upload
from .embedding import Embedding as Embedding
from .chat_model import ChatModel as ChatModel
from .completion import Completion as Completion
from .moderation import Moderation as Moderation
from .audio_model import AudioModel as AudioModel
from .batch_error import BatchError as BatchError
from .file_object import FileObject as FileObject
from .image_model import ImageModel as ImageModel
from .file_content import FileContent as FileContent
from .file_deleted import FileDeleted as FileDeleted
from .model_deleted import ModelDeleted as ModelDeleted
from .images_response import ImagesResponse as ImagesResponse
from .completion_usage import CompletionUsage as CompletionUsage
from .file_list_params import FileListParams as FileListParams
from .moderation_model import ModerationModel as ModerationModel
from .batch_list_params import BatchListParams as BatchListParams
from .completion_choice import CompletionChoice as CompletionChoice
from .image_edit_params import ImageEditParams as ImageEditParams
from .file_create_params import FileCreateParams as FileCreateParams
from .batch_create_params import BatchCreateParams as BatchCreateParams
from .batch_request_counts import BatchRequestCounts as BatchRequestCounts
from .upload_create_params import UploadCreateParams as UploadCreateParams
from .image_generate_params import ImageGenerateParams as ImageGenerateParams
from .upload_complete_params import UploadCompleteParams as UploadCompleteParams
from .embedding_create_params import EmbeddingCreateParams as EmbeddingCreateParams
from .completion_create_params import CompletionCreateParams as CompletionCreateParams
from .moderation_create_params import ModerationCreateParams as ModerationCreateParams
from .create_embedding_response import CreateEmbeddingResponse as CreateEmbeddingResponse
from .moderation_create_response import ModerationCreateResponse as ModerationCreateResponse
from .image_create_variation_params import ImageCreateVariationParams as ImageCreateVariationParams



from unbound import Message, Unbound
from typing import Optional, Union, List, Any, Mapping, cast, Sequence
from unbound.api_resources.utils import UnboundResponse

from unbound.llms.llama_index.utils import (
    IMPORT_ERROR_MESSAGE,
    is_chat_model,
    modelname_to_contextsize,
)

try:
    from llama_index.llms.custom import CustomLLM
    from llama_index.bridge.pydantic import PrivateAttr
    from llama_index.llms.base import (
        ChatMessage,
        ChatResponse,
        ChatResponseGen,
        CompletionResponse,
        CompletionResponseGen,
        LLMMetadata,
        llm_chat_callback,
        llm_completion_callback,
    )
except ImportError as exc:
    raise ImportError(IMPORT_ERROR_MESSAGE) from exc


class UnboundLLM(CustomLLM):
    """_summary_.

    Args:
        LLM (_type_): _description_
    """

    _client: Any = PrivateAttr()
    model: str = ""

    api_key: Optional[str] = None
    base_url: Optional[str] = None
    virtual_key: Optional[str] = None
    config: Optional[Union[Mapping, str]] = None
    provider: Optional[str] = None
    trace_id: Optional[str] = None
    custom_metadata: Optional[str] = None

    def __init__(
        self,
        *,
        api_key: Optional[str] = None,
        base_url: Optional[str] = None,
        virtual_key: Optional[str] = None,
        config: Optional[Union[Mapping, str]] = None,
        provider: Optional[str] = None,
        trace_id: Optional[str] = None,
        custom_metadata: Optional[str] = None,
        **kwargs,
    ) -> None:
        """
        Initialize a Unbound instance.

        Args:
            api_key (Optional[str]): The API key to authenticate with Unbound.
            base_url (Optional[str]): The Base url to the self hosted rubeus \
                (the opensource version of unbound) or any other self hosted server.
        """
        super().__init__(
            base_url=base_url,
            api_key=api_key,
        )
        self._client = Unbound(
            api_key=api_key,
            base_url=base_url,
            virtual_key=virtual_key,
            config=config,
            provider=provider,
            trace_id=trace_id,
            metadata=custom_metadata,
            **kwargs,
        )
        self.model = ""

    @property
    def metadata(self) -> LLMMetadata:
        """LLM metadata."""
        try:
            from llama_index.llms.base import (
                LLMMetadata,
            )
        except ImportError as exc:
            raise ImportError(IMPORT_ERROR_MESSAGE) from exc
        return LLMMetadata(
            _context_window=modelname_to_contextsize(self.model) if self.model else 0,
            is_chat_model=is_chat_model(self.model),
            model_name=self.model,
        )

    @llm_completion_callback()
    def complete(self, prompt: str, **kwargs: Any) -> CompletionResponse:
        """Completion endpoint for LLM."""
        complete_fn = self._complete
        return complete_fn(prompt, **kwargs)

    @llm_chat_callback()
    def chat(self, messages: Sequence[ChatMessage], **kwargs: Any) -> ChatResponse:
        chat_fn = self._chat
        return chat_fn(messages, **kwargs)

    @llm_completion_callback()
    def stream_complete(self, prompt: str, **kwargs: Any) -> CompletionResponseGen:
        """Completion endpoint for LLM."""
        complete_fn = self._stream_complete
        return complete_fn(prompt, **kwargs)

    @llm_chat_callback()
    def stream_chat(
        self, messages: Sequence[ChatMessage], **kwargs: Any
    ) -> ChatResponseGen:
        stream_chat_fn = self._stream_chat
        return stream_chat_fn(messages, **kwargs)

    def _chat(self, messages: Sequence[ChatMessage], **kwargs: Any) -> ChatResponse:
        _messages = cast(
            List[Message],
            [{"role": i.role.value, "content": i.content} for i in messages],
        )
        response = self._client.chat.completions.create(messages=_messages, **kwargs)
        self.model = self._get_model(response)

        message = response.choices[0].message
        return ChatResponse(message=message, raw=response)

    def _complete(self, prompt: str, **kwargs: Any) -> CompletionResponse:
        response = self._client.completions.create(prompt=prompt, **kwargs)
        text = response.choices[0].text
        return CompletionResponse(text=text, raw=response)

    def _stream_chat(
        self, messages: Sequence[ChatMessage], **kwargs: Any
    ) -> ChatResponseGen:
        _messages = cast(
            List[Message],
            [{"role": i.role.value, "content": i.content} for i in messages],
        )
        response = self._client.chat.completions.create(
            messages=_messages, stream=True, **kwargs
        )

        def gen() -> ChatResponseGen:
            content = ""
            function_call: Optional[dict] = {}
            for resp in response:
                if resp.choices is None:
                    continue
                delta = resp.choices[0].delta
                role = delta.get("role", "assistant")
                content_delta = delta.get("content", "") or ""
                content += content_delta

                function_call_delta = delta.get("function_call", None)
                if function_call_delta is not None:
                    if function_call is None:
                        function_call = function_call_delta
                        # ensure we do not add a blank function call
                        if (
                            function_call
                            and function_call.get("function_name", "") is None
                        ):
                            del function_call["function_name"]
                    else:
                        function_call["arguments"] += function_call_delta["arguments"]

                additional_kwargs = {}
                if function_call is not None:
                    additional_kwargs["function_call"] = function_call

                yield ChatResponse(
                    message=ChatMessage(
                        role=role,
                        content=content,
                        additional_kwargs=additional_kwargs,
                    ),
                    delta=content_delta,
                    raw=resp,
                )

        return gen()

    def _stream_complete(self, prompt: str, **kwargs: Any) -> CompletionResponseGen:
        response = self._client.completions.create(prompt=prompt, stream=True, **kwargs)

        def gen() -> CompletionResponseGen:
            text = ""
            for resp in response:
                delta = resp.choices[0].text or ""
                text += delta
                yield CompletionResponse(
                    delta=delta,
                    text=text,
                    raw=resp,
                )

        return gen()

    @property
    def _is_chat_model(self) -> bool:
        """Check if a given model is a chat-based language model.

        Returns:
            bool: True if the provided model is a chat-based language model,
            False otherwise.
        """
        return is_chat_model(self.model or "")

    def _get_model(self, response: UnboundResponse) -> str:
        return response.model



import json
from typing import Dict, Optional
import httpx
from .utils import parse_headers
from pydantic import BaseModel, PrivateAttr


__all__ = ["Thread", "ThreadDeleted"]


class Thread(BaseModel, extra="allow"):
    id: Optional[str]
    created_at: Optional[int]
    metadata: Optional[object] = None
    object: Optional[str]
    _headers: Optional[httpx.Headers] = PrivateAttr()

    def __str__(self):
        del self._headers
        return json.dumps(self.dict(), indent=4)

    def get_headers(self) -> Optional[Dict[str, str]]:
        return parse_headers(self._headers)


class ThreadDeleted(BaseModel, extra="allow"):
    id: Optional[str]
    object: Optional[str]
    deleted: Optional[bool]
    _headers: Optional[httpx.Headers] = PrivateAttr()

    def __str__(self):
        del self._headers
        return json.dumps(self.dict(), indent=4)

    def get_headers(self) -> Optional[Dict[str, str]]:
        return parse_headers(self._headers)



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from __future__ import annotations

from typing_extensions import Literal, Required, TypedDict

__all__ = ["FileSearchToolParam", "FileSearch"]


class FileSearch(TypedDict, total=False):
    max_num_results: int
    """The maximum number of results the file search tool should output.

    The default is 20 for `gpt-4*` models and 5 for `gpt-3.5-turbo`. This number
    should be between 1 and 50 inclusive.

    Note that the file search tool may output fewer than `max_num_results` results.
    See the
    [file search tool documentation](https://platform.openai.com/docs/assistants/tools/file-search/number-of-chunks-returned)
    for more information.
    """


class FileSearchToolParam(TypedDict, total=False):
    type: Required[Literal["file_search"]]
    """The type of tool being defined: `file_search`"""

    file_search: FileSearch
    """Overrides for the file search tool."""



from __future__ import annotations

import io
from typing import Callable
from typing_extensions import override


class CancelledError(Exception):
    def __init__(self, msg: str) -> None:
        self.msg = msg
        super().__init__(msg)

    @override
    def __str__(self) -> str:
        return self.msg

    __repr__ = __str__


class BufferReader(io.BytesIO):
    def __init__(self, buf: bytes = b"", desc: str | None = None) -> None:
        super().__init__(buf)
        self._len = len(buf)
        self._progress = 0
        self._callback = progress(len(buf), desc=desc)

    def __len__(self) -> int:
        return self._len

    @override
    def read(self, n: int | None = -1) -> bytes:
        chunk = io.BytesIO.read(self, n)
        self._progress += len(chunk)

        try:
            self._callback(self._progress)
        except Exception as e:  # catches exception from the callback
            raise CancelledError("The upload was cancelled: {}".format(e)) from e

        return chunk


def progress(total: float, desc: str | None) -> Callable[[float], None]:
    import tqdm

    meter = tqdm.tqdm(total=total, unit_scale=True, desc=desc)

    def incr(progress: float) -> None:
        meter.n = progress
        if progress == total:
            meter.close()
        else:
            meter.refresh()

    return incr


def MB(i: int) -> int:
    return int(i // 1024**2)



import json
from typing import AsyncIterator, Iterator, Optional, Union
from unbound.api_resources.client import AsyncUnbound, Unbound
from ..._vendor.openai._types import NotGiven, NOT_GIVEN

# from unbound.api_resources.utils import TextCompletionChunk
from unbound.api_resources.types.complete_type import (
    TextCompletion,
    TextCompletionChunk,
)
from unbound.api_resources.apis.api_resource import APIResource, AsyncAPIResource


class Completion(APIResource):
    def __init__(self, client: Unbound) -> None:
        super().__init__(client)
        self.openai_client = client.openai_client
        self.client = client

    def stream_create(  # type: ignore[return]
        self, model, prompt, stream, temperature, max_tokens, top_p, **kwargs
    ) -> Union[TextCompletion, Iterator[TextCompletionChunk]]:
        with self.openai_client.with_streaming_response.completions.create(
            model=model,
            prompt=prompt,
            stream=stream,
            temperature=temperature,
            max_tokens=max_tokens,
            top_p=top_p,
            extra_body=kwargs,
        ) as response:
            for line in response.iter_lines():
                json_string = line.replace("data: ", "")
                json_string = json_string.strip().rstrip("\n")
                if json_string == "":
                    continue
                elif json_string == "[DONE]":
                    break
                elif json_string != "":
                    json_data = json.loads(json_string)
                    json_data = TextCompletionChunk(**json_data)
                    yield json_data
                else:
                    return ""

    def normal_create(
        self, model, prompt, stream, temperature, max_tokens, top_p, **kwargs
    ) -> TextCompletion:
        response = self.openai_client.with_raw_response.completions.create(
            model=model,
            prompt=prompt,
            stream=stream,
            temperature=temperature,
            max_tokens=max_tokens,
            top_p=top_p,
            extra_body=kwargs,
        )
        data = TextCompletion(**json.loads(response.text))
        data._headers = response.headers
        return data

    def create(
        self,
        *,
        model: Optional[str] = "unbound-default",
        prompt: Optional[str] = None,
        stream: Union[bool, NotGiven] = NOT_GIVEN,
        temperature: Union[float, NotGiven] = NOT_GIVEN,
        max_tokens: Union[int, NotGiven] = NOT_GIVEN,
        top_p: Union[bool, NotGiven] = NOT_GIVEN,
        **kwargs,
    ) -> Union[TextCompletion, Iterator[TextCompletionChunk]]:
        if stream is True:
            return self.stream_create(
                model=model,
                prompt=prompt,
                stream=stream,
                temperature=temperature,
                max_tokens=max_tokens,
                top_p=top_p,
                **kwargs,
            )
        else:
            return self.normal_create(
                model=model,
                prompt=prompt,
                stream=stream,
                temperature=temperature,
                max_tokens=max_tokens,
                top_p=top_p,
                **kwargs,
            )


class AsyncCompletion(AsyncAPIResource):
    def __init__(self, client: AsyncUnbound) -> None:
        super().__init__(client)
        self.openai_client = client.openai_client

    async def stream_create(
        self, model, prompt, stream, temperature, max_tokens, top_p, **kwargs
    ) -> Union[TextCompletion, AsyncIterator[TextCompletionChunk]]:
        async with self.openai_client.with_streaming_response.completions.create(
            model=model,
            prompt=prompt,
            stream=stream,
            temperature=temperature,
            max_tokens=max_tokens,
            top_p=top_p,
            extra_body=kwargs,
        ) as response:
            async for line in response.iter_lines():
                json_string = line.replace("data: ", "")
                json_string = json_string.strip().rstrip("\n")
                if json_string == "":
                    continue
                elif json_string == "[DONE]":
                    break
                elif json_string != "":
                    json_data = json.loads(json_string)
                    json_data = TextCompletionChunk(**json_data)
                    yield json_data
                else:
                    pass

    async def normal_create(
        self, model, prompt, stream, temperature, max_tokens, top_p, **kwargs
    ) -> TextCompletion:
        response = await self.openai_client.with_raw_response.completions.create(
            model=model,
            prompt=prompt,
            stream=stream,
            temperature=temperature,
            max_tokens=max_tokens,
            top_p=top_p,
            extra_body=kwargs,
        )
        data = TextCompletion(**json.loads(response.text))
        data._headers = response.headers
        return data

    async def create(
        self,
        *,
        model: Optional[str] = "unbound-default",
        prompt: Optional[str] = None,
        stream: Union[bool, NotGiven] = NOT_GIVEN,
        temperature: Union[float, NotGiven] = NOT_GIVEN,
        max_tokens: Union[int, NotGiven] = NOT_GIVEN,
        top_p: Union[bool, NotGiven] = NOT_GIVEN,
        **kwargs,
    ) -> Union[TextCompletion, AsyncIterator[TextCompletionChunk]]:
        if stream is True:
            return self.stream_create(
                model=model,
                prompt=prompt,
                stream=stream,
                temperature=temperature,
                max_tokens=max_tokens,
                top_p=top_p,
                **kwargs,
            )
        else:
            return await self.normal_create(
                model=model,
                prompt=prompt,
                stream=stream,
                temperature=temperature,
                max_tokens=max_tokens,
                top_p=top_p,
                **kwargs,
            )



""""""
from .apis import (
    Completion,
    AsyncCompletion,
    ChatCompletion,
    AsyncChatCompletion,
    Generations,
    AsyncGenerations,
    Prompts,
    AsyncPrompts,
    Feedback,
    AsyncFeedback,
    createHeaders,
    Images,
    AsyncImages,
    Assistants,
    AsyncAssistants,
    Threads,
    AsyncThreads,
    Messages,
    AsyncMessages,
    MainFiles,
    AsyncMainFiles,
    Models,
    AsyncModels,
    Runs,
    AsyncRuns,
    Steps,
    AsyncSteps,
    Moderations,
    AsyncModerations,
    Audio,
    Transcriptions,
    Translations,
    Speech,
    AsyncAudio,
    AsyncTranscriptions,
    AsyncTranslations,
    AsyncSpeech,
    Batches,
    AsyncBatches,
    FineTuning,
    Jobs,
    Checkpoints,
    AsyncFineTuning,
    AsyncJobs,
    AsyncCheckpoints,
    VectorStores,
    VectorFiles,
    VectorFileBatches,
    AsyncVectorStores,
    AsyncVectorFiles,
    AsyncVectorFileBatches,
    Admin,
    Users,
    Invites,
    Workspaces,
    WorkspacesUsers,
    AsyncAdmin,
    AsyncUsers,
    AsyncInvites,
    AsyncWorkspaces,
    AsyncWorkspacesUsers,
    BetaChat,
    AsyncBetaChat,
    BetaCompletions,
    AsyncBetaCompletions,
    Uploads,
    AsyncUploads,
    Parts,
    AsyncParts,
    Configs,
    AsyncConfigs,
    ApiKeys,
    AsyncApiKeys,
    VirtualKeys,
    AsyncVirtualKeys,
    Logs,
    AsyncLogs,
)
from .utils import (
    Modes,
    ModesLiteral,
    LLMOptions,
    ProviderTypes,
    ProviderTypesLiteral,
    CacheType,
    CacheLiteral,
    Message,
    UnboundResponse,
    Params,
    Config,
    RetrySettings,
)
from .client import Unbound, AsyncUnbound

from unbound.version import VERSION

__version__ = VERSION
__all__ = [
    "LLMOptions",
    "Modes",
    "UnboundResponse",
    "ModesLiteral",
    "ProviderTypes",
    "ProviderTypesLiteral",
    "CacheType",
    "CacheLiteral",
    "Message",
    "Completion",
    "AsyncCompletion",
    "Params",
    "Config",
    "RetrySettings",
    "ChatCompletion",
    "AsyncChatCompletion",
    "Generations",
    "AsyncGenerations",
    "Prompts",
    "AsyncPrompts",
    "Feedback",
    "AsyncFeedback",
    "createHeaders",
    "Unbound",
    "AsyncUnbound",
    "Images",
    "AsyncImages",
    "Assistants",
    "AsyncAssistants",
    "Threads",
    "AsyncThreads",
    "Messages",
    "AsyncMessages",
    "MainFiles",
    "AsyncMainFiles",
    "Models",
    "AsyncModels",
    "Runs",
    "AsyncRuns",
    "Steps",
    "AsyncSteps",
    "Moderations",
    "AsyncModerations",
    "Audio",
    "Transcriptions",
    "Translations",
    "Speech",
    "AsyncAudio",
    "AsyncTranscriptions",
    "AsyncTranslations",
    "AsyncSpeech",
    "Batches",
    "AsyncBatches",
    "FineTuning",
    "Jobs",
    "Checkpoints",
    "AsyncFineTuning",
    "AsyncJobs",
    "AsyncCheckpoints",
    "VectorStores",
    "VectorFiles",
    "VectorFileBatches",
    "AsyncVectorStores",
    "AsyncVectorFiles",
    "AsyncVectorFileBatches",
    "Admin",
    "Users",
    "Invites",
    "Workspaces",
    "WorkspacesUsers",
    "AsyncAdmin",
    "AsyncUsers",
    "AsyncInvites",
    "AsyncWorkspaces",
    "AsyncWorkspacesUsers",
    "BetaChat",
    "AsyncBetaChat",
    "BetaCompletions",
    "AsyncBetaCompletions",
    "Uploads",
    "Parts",
    "AsyncUploads",
    "AsyncParts",
    "Configs",
    "AsyncConfigs",
    "ApiKeys",
    "AsyncApiKeys",
    "VirtualKeys",
    "AsyncVirtualKeys",
    "Logs",
    "AsyncLogs",
]



from typing import Any, Dict, List, Literal, Optional, Union
from unbound.api_resources.base_client import APIClient, AsyncAPIClient
from urllib.parse import urlencode
from unbound.api_resources.apis.api_resource import APIResource, AsyncAPIResource
from unbound.api_resources.types.user_invite_type import (
    UserInviteResponse,
    UserInviteRetrieveAllResponse,
    UserInviteRetrieveResponse,
    UserRetrieveAllResponse,
    UserRetrieveResponse,
    WorkspaceMemberGetResponse,
    WorkspaceMemberListResponse,
    WorkspacesAddResponse,
    WorkspacesGetResponse,
    WorkspacesListResponse,
    WorkspacesUpdateResponse,
)
from unbound.api_resources.utils import GenericResponse
from unbound.api_resources.utils import UnboundApiPaths


class Admin(APIResource):
    def __init__(self, client: APIClient) -> None:
        super().__init__(client)
        self.users = Users(client)
        self.workspaces = Workspaces(client)


class Users(APIResource):
    def __init__(self, client: APIClient) -> None:
        super().__init__(client)
        self.invites = Invites(client)

    def retrieve(self, *, user_id: str) -> UserRetrieveResponse:
        return self._get(
            f"{UnboundApiPaths.USER_API}/{user_id}",
            params=None,
            body=None,
            cast_to=UserRetrieveResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    def list(
        self,
        *,
        page_size: Optional[Union[int, str]] = None,
        current_page: Optional[int] = 0,
        email: Optional[str] = None,
        role: Optional[str] = None,
    ) -> UserRetrieveAllResponse:
        query = {
            "pageSize": page_size,
            "currentPage": current_page,
            "email": email,
            "role": role,
        }
        filtered_query = {k: v for k, v in query.items() if v is not None}
        query_string = urlencode(filtered_query)
        return self._get(
            f"{UnboundApiPaths.USER_API}?{query_string}",
            params=None,
            body=None,
            cast_to=UserRetrieveAllResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    def update(self, *, user_id: str, role: str) -> Any:
        body = {"role": role}
        return self._put(
            f"{UnboundApiPaths.USER_API}/{user_id}",
            body=body,
            params=None,
            cast_to=GenericResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    def delete(self, *, user_id: str) -> Any:
        return self._delete(
            f"{UnboundApiPaths.USER_API}/{user_id}",
            params=None,
            body=None,
            cast_to=GenericResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )


class Invites(APIResource):
    def __init__(self, client: APIClient) -> None:
        super().__init__(client)

    def create(
        self,
        *,
        email: Optional[str] = None,
        role: Optional[str] = None,
        workspaces: Optional[List[Dict[str, Any]]] = None,
        workspace_api_key_details: Optional[Dict[str, Any]] = None,
    ) -> UserInviteResponse:
        body = {
            "email": email,
            "role": role,
            "workspaces": workspaces,
            "workspace_api_key_details": workspace_api_key_details,
        }
        return self._post(
            f"{UnboundApiPaths.INVITE_API}",
            body=body,
            params=None,
            cast_to=UserInviteResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    def retrieve(self, *, invite_id: str) -> UserInviteRetrieveResponse:
        return self._get(
            f"{UnboundApiPaths.INVITE_API}/{invite_id}",
            params=None,
            body=None,
            cast_to=UserInviteRetrieveResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    def list(
        self,
        *,
        page_size: Optional[Union[int, str]] = None,
        current_page: Optional[int] = 0,
        email: Optional[str] = None,
        role: Optional[str] = None,
        status: Optional[str] = None,
    ) -> UserInviteRetrieveAllResponse:
        query = {
            "pageSize": page_size,
            "currentPage": current_page,
            "email": email,
            "role": role,
            "status": status,
        }
        filtered_query = {k: v for k, v in query.items() if v is not None}
        query_string = urlencode(filtered_query)
        return self._get(
            f"{UnboundApiPaths.INVITE_API}?{query_string}",
            params=None,
            body=None,
            cast_to=UserInviteRetrieveAllResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    def delete(self, *, invite_id: str) -> Any:
        return self._delete(
            f"{UnboundApiPaths.INVITE_API}/{invite_id}",
            params=None,
            body=None,
            cast_to=GenericResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )


class Workspaces(APIResource):
    def __init__(self, client: APIClient) -> None:
        super().__init__(client)
        self.users = WorkspacesUsers(client)

    def create(
        self,
        *,
        name: Optional[str] = None,
        description: Optional[str] = None,
        defaults: Optional[Dict[str, Any]] = None,
        users: Optional[List[str]] = None,
    ) -> WorkspacesAddResponse:
        body = {
            "name": name,
            "description": description,
            "defaults": defaults,
            "users": users,
        }
        return self._post(
            f"{UnboundApiPaths.WORKSPACE_API}",
            body=body,
            params=None,
            cast_to=WorkspacesAddResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    def retrieve(self, *, workspace_id: str) -> WorkspacesGetResponse:
        return self._get(
            f"{UnboundApiPaths.WORKSPACE_API}/{workspace_id}",
            params=None,
            body=None,
            cast_to=WorkspacesGetResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    def list(
        self,
        *,
        page_size: Optional[Union[int, str]] = None,
        current_page: Optional[int] = 0,
    ) -> WorkspacesListResponse:
        query = {
            "pageSize": page_size,
            "currentPage": current_page,
        }
        filtered_query = {k: v for k, v in query.items() if v is not None}
        query_string = urlencode(filtered_query)
        return self._get(
            f"{UnboundApiPaths.WORKSPACE_API}?{query_string}",
            params=None,
            body=None,
            cast_to=WorkspacesListResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    def update(
        self,
        *,
        workspace_id: Optional[str] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        defaults: Optional[Dict[str, Any]] = None,
    ) -> WorkspacesUpdateResponse:
        body = {
            "name": name,
            "description": description,
            "defaults": defaults,
        }
        return self._put(
            f"{UnboundApiPaths.WORKSPACE_API}/{workspace_id}",
            body=body,
            params=None,
            cast_to=WorkspacesUpdateResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    def delete(
        self,
        *,
        workspace_id: Optional[str],
        name: Optional[str] = None,
    ) -> Any:
        body = {
            "workspace_id": workspace_id,
            "name": name,
        }
        return self._delete(
            f"{UnboundApiPaths.WORKSPACE_API}/{workspace_id}",
            params=None,
            body=body,
            cast_to=GenericResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )


class WorkspacesUsers(APIResource):
    def __init__(self, client: APIClient) -> None:
        super().__init__(client)

    def create(
        self,
        *,
        workspace_id: Optional[str] = None,
        users: Optional[List[Dict[str, str]]] = None,
    ) -> Any:
        body = {
            "workspace_id": workspace_id,
            "users": users,
        }
        return self._post(
            f"{UnboundApiPaths.WORKSPACE_API}/{workspace_id}/users",
            body=body,
            params=None,
            cast_to=GenericResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    def retrieve(
        self,
        *,
        workspace_id: Optional[str] = None,
        user_id: Optional[str] = None,
    ) -> WorkspaceMemberGetResponse:
        return self._get(
            f"{UnboundApiPaths.WORKSPACE_API}/{workspace_id}/users/{user_id}",
            params=None,
            body=None,
            cast_to=WorkspaceMemberGetResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    def list(
        self,
        *,
        workspace_id: Optional[str] = None,
        page_size: Optional[Union[int, str]] = None,
        current_page: Optional[int] = 0,
        email: Optional[str] = None,
        role: Optional[Union[Literal["admin", "manager", "member"], str]] = None,
    ) -> Any:
        query = {
            "page_size": page_size,
            "current_page": current_page,
            "email": email,
            "role": role,
        }
        filtered_query = {k: v for k, v in query.items() if v is not None}
        query_string = urlencode(filtered_query)
        return self._get(
            f"{UnboundApiPaths.WORKSPACE_API}/{workspace_id}/users?{query_string}",
            params=None,
            body=None,
            cast_to=GenericResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    def update(
        self,
        *,
        workspace_id: Optional[str] = None,
        user_id: Optional[str] = None,
        role: Optional[Union[Literal["admin", "manager", "member"], str]] = None,
    ) -> Any:
        body = {
            "user_id": user_id,
            "role": role,
        }
        return self._put(
            f"{UnboundApiPaths.WORKSPACE_API}/{workspace_id}/users/{user_id}",
            body=body,
            params=None,
            cast_to=GenericResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    def delete(
        self,
        *,
        workspace_id: Optional[str],
        user_id: Optional[str] = None,
    ) -> Any:
        return self._delete(
            f"{UnboundApiPaths.WORKSPACE_API}/{workspace_id}/users/{user_id}",
            params=None,
            body=None,
            cast_to=GenericResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )


class AsyncAdmin(AsyncAPIResource):
    def __init__(self, client: AsyncAPIClient) -> None:
        super().__init__(client)
        self.users = AsyncUsers(client)
        self.workspaces = AsyncWorkspaces(client)


class AsyncUsers(AsyncAPIResource):
    def __init__(self, client: AsyncAPIClient) -> None:
        super().__init__(client)
        self.invites = AsyncInvites(client)

    async def retrieve(self, *, user_id: str) -> UserRetrieveResponse:
        return await self._get(
            f"{UnboundApiPaths.USER_API}/{user_id}",
            params=None,
            body=None,
            cast_to=UserRetrieveResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    async def list(
        self,
        *,
        page_size: Optional[Union[int, str]] = None,
        current_page: Optional[int] = 0,
        email: Optional[str] = None,
        role: Optional[str] = None,
    ) -> UserRetrieveAllResponse:
        query = {
            "pageSize": page_size,
            "currentPage": current_page,
            "email": email,
            "role": role,
        }
        filtered_query = {k: v for k, v in query.items() if v is not None}
        query_string = urlencode(filtered_query)
        return await self._get(
            f"{UnboundApiPaths.USER_API}?{query_string}",
            params=None,
            body=None,
            cast_to=UserRetrieveAllResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    async def update(self, *, user_id: str, role: str) -> Any:
        body = {"role": role}
        return await self._put(
            f"{UnboundApiPaths.USER_API}/{user_id}",
            body=body,
            params=None,
            cast_to=GenericResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    async def delete(self, *, user_id: str) -> Any:
        return await self._delete(
            f"{UnboundApiPaths.USER_API}/{user_id}",
            params=None,
            body=None,
            cast_to=GenericResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )


class AsyncInvites(AsyncAPIResource):
    def __init__(self, client: AsyncAPIClient) -> None:
        super().__init__(client)

    async def create(
        self,
        *,
        email: Optional[str] = None,
        role: Optional[str] = None,
        workspaces: Optional[List[Dict[str, Any]]] = None,
        workspace_api_key_details: Optional[Dict[str, Any]] = None,
    ) -> UserInviteResponse:
        body = {
            "email": email,
            "role": role,
            "workspaces": workspaces,
            "workspace_api_key_details": workspace_api_key_details,
        }
        return await self._post(
            f"{UnboundApiPaths.INVITE_API}",
            body=body,
            params=None,
            cast_to=UserInviteResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    async def retrieve(self, *, invite_id: str) -> UserInviteRetrieveResponse:
        return await self._get(
            f"{UnboundApiPaths.INVITE_API}/{invite_id}",
            params=None,
            body=None,
            cast_to=UserInviteRetrieveResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    async def list(
        self,
        *,
        page_size: Optional[Union[int, str]] = None,
        current_page: Optional[int] = 0,
        email: Optional[str] = None,
        role: Optional[str] = None,
        status: Optional[str] = None,
    ) -> UserInviteRetrieveAllResponse:
        query = {
            "pageSize": page_size,
            "currentPage": current_page,
            "email": email,
            "role": role,
            "status": status,
        }
        filtered_query = {k: v for k, v in query.items() if v is not None}
        query_string = urlencode(filtered_query)
        return await self._get(
            f"{UnboundApiPaths.INVITE_API}?{query_string}",
            params=None,
            body=None,
            cast_to=UserInviteRetrieveAllResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    async def delete(self, *, invite_id: str) -> Any:
        return await self._delete(
            f"{UnboundApiPaths.INVITE_API}/{invite_id}",
            params=None,
            body=None,
            cast_to=GenericResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )


class AsyncWorkspaces(AsyncAPIResource):
    def __init__(self, client: AsyncAPIClient) -> None:
        super().__init__(client)
        self.users = AsyncWorkspacesUsers(client)

    async def create(
        self,
        *,
        name: Optional[str] = None,
        description: Optional[str] = None,
        defaults: Optional[Dict[str, Any]] = None,
        users: Optional[List[str]] = None,
    ) -> WorkspacesAddResponse:
        body = {
            "name": name,
            "description": description,
            "defaults": defaults,
            "users": users,
        }
        return await self._post(
            f"{UnboundApiPaths.WORKSPACE_API}",
            body=body,
            params=None,
            cast_to=WorkspacesAddResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    async def retrieve(self, *, workspace_id: str) -> WorkspacesGetResponse:
        return await self._get(
            f"{UnboundApiPaths.WORKSPACE_API}/{workspace_id}",
            params=None,
            body=None,
            cast_to=WorkspacesGetResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    async def list(
        self,
        *,
        page_size: Optional[Union[int, str]] = None,
        current_page: Optional[int] = 0,
    ) -> WorkspacesListResponse:
        query = {
            "pageSize": page_size,
            "currentPage": current_page,
        }
        filtered_query = {k: v for k, v in query.items() if v is not None}
        query_string = urlencode(filtered_query)
        return await self._get(
            f"{UnboundApiPaths.WORKSPACE_API}?{query_string}",
            params=None,
            body=None,
            cast_to=WorkspacesListResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    async def update(
        self,
        *,
        workspace_id: Optional[str] = None,
        name: Optional[str] = None,
        description: Optional[str] = None,
        defaults: Optional[Dict[str, Any]] = None,
    ) -> WorkspacesUpdateResponse:
        body = {
            "name": name,
            "description": description,
            "defaults": defaults,
        }
        return await self._put(
            f"{UnboundApiPaths.WORKSPACE_API}/{workspace_id}",
            body=body,
            params=None,
            cast_to=WorkspacesUpdateResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    async def delete(
        self,
        *,
        workspace_id: Optional[str],
        name: Optional[str] = None,
    ) -> Any:
        body = {
            "workspace_id": workspace_id,
            "name": name,
        }
        return await self._delete(
            f"{UnboundApiPaths.WORKSPACE_API}/{workspace_id}",
            params=None,
            body=body,
            cast_to=GenericResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )


class AsyncWorkspacesUsers(AsyncAPIResource):
    def __init__(self, client: AsyncAPIClient) -> None:
        super().__init__(client)

    async def create(
        self,
        *,
        workspace_id: Optional[str] = None,
        users: Optional[List[Dict[str, str]]] = None,
    ) -> Any:
        body = {
            "workspace_id": workspace_id,
            "users": users,
        }
        return await self._post(
            f"{UnboundApiPaths.WORKSPACE_API}/{workspace_id}/users",
            body=body,
            params=None,
            cast_to=GenericResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    async def retrieve(
        self,
        *,
        workspace_id: Optional[str] = None,
        user_id: Optional[str] = None,
    ) -> WorkspaceMemberGetResponse:
        return await self._get(
            f"{UnboundApiPaths.WORKSPACE_API}/{workspace_id}/users/{user_id}",
            params=None,
            body=None,
            cast_to=WorkspaceMemberGetResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    async def list(
        self,
        *,
        workspace_id: Optional[str] = None,
        page_size: Optional[Union[int, str]] = None,
        current_page: Optional[int] = 0,
        email: Optional[str] = None,
        role: Optional[Union[Literal["admin", "manager", "member"], str]] = None,
    ) -> WorkspaceMemberListResponse:
        query = {
            "page_size": page_size,
            "current_page": current_page,
            "email": email,
            "role": role,
        }
        filtered_query = {k: v for k, v in query.items() if v is not None}
        query_string = urlencode(filtered_query)
        return await self._get(
            f"{UnboundApiPaths.WORKSPACE_API}/{workspace_id}/users?{query_string}",
            params=None,
            body=None,
            cast_to=WorkspaceMemberListResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    async def update(
        self,
        *,
        workspace_id: Optional[str] = None,
        user_id: Optional[str] = None,
        role: Optional[Union[Literal["admin", "manager", "member"], str]] = None,
    ) -> Any:
        body = {
            "user_id": user_id,
            "role": role,
        }
        return await self._put(
            f"{UnboundApiPaths.WORKSPACE_API}/{workspace_id}/users/{user_id}",
            body=body,
            params=None,
            cast_to=GenericResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )

    async def delete(
        self,
        *,
        workspace_id: Optional[str],
        user_id: Optional[str] = None,
    ) -> Any:
        return await self._delete(
            f"{UnboundApiPaths.WORKSPACE_API}/{workspace_id}/users/{user_id}",
            params=None,
            body=None,
            cast_to=GenericResponse,
            stream=False,
            stream_cls=None,
            headers={},
        )



import json
from typing import Dict, Optional
import httpx
from .utils import parse_headers
from typing import List, Any
from pydantic import BaseModel, PrivateAttr

__all__ = ["FileObject", "FileList", "FileDeleted"]


class FileObject(BaseModel):
    id: Optional[str]
    bytes: Optional[int]
    created_at: Optional[int]
    filename: Optional[str]
    object: Optional[str]
    purpose: Optional[str]
    status: Optional[str]
    status_details: Optional[str] = None
    _headers: Optional[httpx.Headers] = PrivateAttr()

    def __str__(self):
        del self._headers
        return json.dumps(self.dict(), indent=4)

    def __getitem__(self, key):
        return getattr(self, key, None)

    def get(self, key: str, default: Optional[Any] = None):
        return getattr(self, key, None) or default

    def get_headers(self) -> Optional[Dict[str, str]]:
        return parse_headers(self._headers)


class FileList(BaseModel):
    object: Optional[str]
    data: List[FileObject]
    _headers: Optional[httpx.Headers] = PrivateAttr()

    def __str__(self):
        del self._headers
        return json.dumps(self.dict(), indent=4)

    def __getitem__(self, key):
        return getattr(self, key, None)

    def get(self, key: str, default: Optional[Any] = None):
        return getattr(self, key, None) or default

    def get_headers(self) -> Optional[Dict[str, str]]:
        return parse_headers(self._headers)


class FileDeleted(BaseModel):
    id: Optional[str]
    deleted: Optional[bool]
    object: Optional[str]
    _headers: Optional[httpx.Headers] = PrivateAttr()

    def __str__(self):
        del self._headers
        return json.dumps(self.dict(), indent=4)

    def __getitem__(self, key):
        return getattr(self, key, None)

    def get(self, key: str, default: Optional[Any] = None):
        return getattr(self, key, None) or default

    def get_headers(self) -> Optional[Dict[str, str]]:
        return parse_headers(self._headers)



import json
from typing import Union, Any
import typing
from unbound.api_resources.apis.api_resource import APIResource, AsyncAPIResource
from unbound.api_resources.client import AsyncUnbound, Unbound
from unbound.api_resources.types.image_type import ImagesResponse
from ..._vendor.openai._types import NotGiven, NOT_GIVEN


class Images(APIResource):
    def __init__(self, client: Unbound) -> None:
        super().__init__(client)
        self.openai_client = client.openai_client

    @typing.no_type_check
    def generate(
        self,
        *,
        prompt: str,
        model: Union[str, NotGiven] = NOT_GIVEN,
        n: Union[int, NotGiven] = NOT_GIVEN,
        quality: Union[str, NotGiven] = NOT_GIVEN,
        response_format: Union[str, NotGiven] = NOT_GIVEN,
        size: Union[str, NotGiven] = NOT_GIVEN,
        user: Union[str, NotGiven] = NOT_GIVEN,
        style: Union[str, NotGiven] = NOT_GIVEN,
        **kwargs
    ) -> ImagesResponse:
        response = self.openai_client.with_raw_response.images.generate(
            prompt=prompt,
            model=model,
            n=n,
            quality=quality,
            response_format=response_format,
            size=size,
            style=style,
            user=user,
            extra_body=kwargs,
        )
        data = ImagesResponse(**json.loads(response.text))
        data._headers = response.headers

        return data

    @typing.no_type_check
    def edit(
        self,
        *,
        prompt: str,
        image,
        mask: Union[Any, NotGiven] = NOT_GIVEN,
        model: Union[str, NotGiven] = NOT_GIVEN,
        n: Union[int, NotGiven] = NOT_GIVEN,
        response_format: Union[str, NotGiven] = NOT_GIVEN,
        size: Union[str, NotGiven] = NOT_GIVEN,
        user: Union[str, NotGiven] = NOT_GIVEN,
        **kwargs
    ) -> ImagesResponse:
        response = self.openai_client.with_raw_response.images.edit(
            prompt=prompt,
            image=image,
            mask=mask,
            model=model,
            n=n,
            response_format=response_format,
            size=size,
            user=user,
            extra_body=kwargs,
        )
        data = ImagesResponse(**json.loads(response.text))
        data._headers = response.headers

        return data

    @typing.no_type_check
    def create_variation(
        self,
        *,
        image,
        n: Union[int, NotGiven] = NOT_GIVEN,
        response_format: Union[str, NotGiven] = NOT_GIVEN,
        size: Union[str, NotGiven] = NOT_GIVEN,
        user: Union[str, NotGiven] = NOT_GIVEN,
        **kwargs
    ) -> ImagesResponse:
        response = self.openai_client.with_raw_response.images.create_variation(
            image=image,
            n=n,
            response_format=response_format,
            size=size,
            user=user,
            extra_body=kwargs,
        )
        data = ImagesResponse(**json.loads(response.text))
        data._headers = response.headers

        return data


class AsyncImages(AsyncAPIResource):
    def __init__(self, client: AsyncUnbound) -> None:
        super().__init__(client)
        self.openai_client = client.openai_client

    @typing.no_type_check
    async def generate(
        self,
        *,
        prompt: str,
        model: Union[str, NotGiven] = NOT_GIVEN,
        n: Union[int, NotGiven] = NOT_GIVEN,
        quality: Union[str, NotGiven] = NOT_GIVEN,
        response_format: Union[str, NotGiven] = NOT_GIVEN,
        size: Union[str, NotGiven] = NOT_GIVEN,
        user: Union[str, NotGiven] = NOT_GIVEN,
        style: Union[str, NotGiven] = NOT_GIVEN,
        **kwargs
    ) -> ImagesResponse:
        response = await self.openai_client.with_raw_response.images.generate(
            prompt=prompt,
            model=model,
            n=n,
            quality=quality,
            response_format=response_format,
            size=size,
            style=style,
            user=user,
            extra_body=kwargs,
        )
        data = ImagesResponse(**json.loads(response.text))
        data._headers = response.headers

        return data

    @typing.no_type_check
    async def edit(
        self,
        *,
        prompt: str,
        image,
        mask: Union[Any, NotGiven] = NOT_GIVEN,
        model: Union[str, NotGiven] = NOT_GIVEN,
        n: Union[int, NotGiven] = NOT_GIVEN,
        response_format: Union[str, NotGiven] = NOT_GIVEN,
        size: Union[str, NotGiven] = NOT_GIVEN,
        user: Union[str, NotGiven] = NOT_GIVEN,
        **kwargs
    ) -> ImagesResponse:
        response = await self.openai_client.with_raw_response.images.edit(
            prompt=prompt,
            image=image,
            mask=mask,
            model=model,
            n=n,
            response_format=response_format,
            size=size,
            user=user,
            extra_body=kwargs,
        )
        data = ImagesResponse(**json.loads(response.text))
        data._headers = response.headers

        return data

    @typing.no_type_check
    async def create_variation(
        self,
        *,
        image,
        n: Union[int, NotGiven] = NOT_GIVEN,
        response_format: Union[str, NotGiven] = NOT_GIVEN,
        size: Union[str, NotGiven] = NOT_GIVEN,
        user: Union[str, NotGiven] = NOT_GIVEN,
        **kwargs
    ) -> ImagesResponse:
        response = await self.openai_client.with_raw_response.images.create_variation(
            image=image,
            n=n,
            response_format=response_format,
            size=size,
            user=user,
            extra_body=kwargs,
        )
        data = ImagesResponse(**json.loads(response.text))
        data._headers = response.headers

        return data



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from typing import Optional
from typing_extensions import Literal

from ...._models import BaseModel

__all__ = ["FileCitationDeltaAnnotation", "FileCitation"]


class FileCitation(BaseModel):
    file_id: Optional[str] = None
    """The ID of the specific File the citation is from."""

    quote: Optional[str] = None
    """The specific quote in the file."""


class FileCitationDeltaAnnotation(BaseModel):
    index: int
    """The index of the annotation in the text content part."""

    type: Literal["file_citation"]
    """Always `file_citation`."""

    end_index: Optional[int] = None

    file_citation: Optional[FileCitation] = None

    start_index: Optional[int] = None

    text: Optional[str] = None
    """The text in the message content that needs to be replaced."""



# File generated from our OpenAPI spec by Stainless. See CONTRIBUTING.md for details.

from .error_object import ErrorObject as ErrorObject
from .function_definition import FunctionDefinition as FunctionDefinition
from .function_parameters import FunctionParameters as FunctionParameters
from .response_format_text import ResponseFormatText as ResponseFormatText
from .response_format_json_object import ResponseFormatJSONObject as ResponseFormatJSONObject
from .response_format_json_schema import ResponseFormatJSONSchema as ResponseFormatJSONSchema
